"""Simulation utilitites."""

from collections.abc import Sequence
from functools import partial
from typing import cast

import jax
import jax.numpy as jnp
import numpy as np
import pandas as pd
import xarray as xr
from jaxtyping import Array
from numpy.typing import NDArray
from scipy.constants import c


def interpolate_xarray(
    xarr: xr.DataArray,
    /,
    *,
    f: Array,
    **kwargs: Array,
) -> dict[str, Array]:
    with jax.ensure_compile_time_eval():
        data = jnp.asarray(xarr)

        # don't use jnp.asarray here as values can be strings!
        params = {k: np.asarray(xarr[k]) for k in xarr.dims}
        strings: dict[str, dict[str, int]] = kwargs.pop("strings", {})  # type: ignore[reportAssignmentType]
        target_name = xarr.dims[-1]
        params.pop(target_name, None)
        strings.pop(target_name, None)  # type: ignore[reportArgumentType]

        # don't use jnp.asarray here as values can be strings!
        targets = {
            str(k): i for i, k in enumerate(np.asarray(xarr.coords[target_name]))
        }

        params["targets"] = np.arange(0, len(targets), 1, dtype=np.uint8)  # type: ignore
        strings = {**strings, "targets": targets}

    S, axs, pos = evaluate_general_corner_model(
        data,
        params,  # type: ignore[reportArgumentType]
        strings,
        **kwargs,
        f=f,
    )
    return {k: S.take(pos["targets"][k], axs["targets"]) for k in targets}


def evaluate_general_corner_model(
    data: Array,
    params: dict[str, Array],
    strings: dict[str, dict[str, int]],
    /,
    **kwargs: Array,
) -> tuple[Array, dict[str, int], dict[str, dict[str, int]]]:
    """Hypercube representation interpolator.

    Args:
        data: the data to interpolate (output of to_hypercube)
        params: the parameter sample points (output of to_hypercube)
        strings: non-interpolatable parameters (like strings)
            with their values (output of to_hypercube)
        **kwargs: the values at which to interpolate the s-parameter.
            These values should be broadcastable into each other.

    Returns:
        the interpolated S-matrix at the values of **kwargs
    """
    given_params = {k: kwargs.get(k, jnp.asarray(v).mean()) for k, v in params.items()}
    given_strings = {
        p: (list(v.values()) if p not in kwargs else [v[vv] for vv in kwargs[p]])
        for p, v in strings.items()
    }
    string_locations = {k: i for i, k in enumerate(given_params) if k in given_strings}
    param_locations = {
        k: i for i, k in enumerate(given_params) if k not in given_strings
    }
    data = jnp.transpose(
        data, axes=[*param_locations.values(), *string_locations.values()]
    )
    given_params = {k: v for k, v in given_params.items() if k not in given_strings}
    num_params = len(given_params)
    data = _downselect(data, list(given_strings.values()))
    param_shape, string_shape = data.shape[:num_params], data.shape[num_params:]
    data = data.reshape(*param_shape, -1)
    given_param_values = [jnp.asarray(v) for v in given_params.values()]
    stacked_params = jnp.stack(jnp.broadcast_arrays(*given_param_values), 0)
    coords = _get_coordinates(
        [v for k, v in params.items() if k in given_params], stacked_params
    )
    result = _map_coordinates(data, coords)
    axs = {k: i for i, k in enumerate(given_strings)}
    rev_strings = {k: {vv: kk for kk, vv in v.items()} for k, v in strings.items()}
    pos = {
        k: {rev_strings[k][vv]: i for i, vv in enumerate(v)}
        for k, v in given_strings.items()
    }
    return result.reshape(*string_shape, *result.shape[1:]), axs, pos


def _downselect(data: Array, idxs_list: list[list[int]]) -> Array:
    for i, idxs in enumerate(reversed(idxs_list), start=1):
        data = data.take(np.asarray(idxs, dtype=int), axis=-i)
    return data


@partial(jax.vmap, in_axes=(-1, None), out_axes=0)
def _map_coordinates(input: Array, coordinates: Sequence[Array]) -> Array:  # noqa: A002
    return jax.scipy.ndimage.map_coordinates(input, coordinates, 1, mode="nearest")


def _get_coordinate(arr1d: Array, value: Array) -> Array:
    return jnp.interp(value, arr1d, jnp.arange(arr1d.shape[0]))


def _get_coordinates(arrs1d: Sequence[Array], values: jnp.ndarray) -> list[Array]:
    # don't use vmap as arrays in arrs1d could have different shapes...
    return [_get_coordinate(a, v) for a, v in zip(arrs1d, values, strict=True)]


def to_hypercube(
    stacked_data: pd.DataFrame, wl_key: str = "wl"
) -> tuple[Array, dict[str, Array], dict[str, dict[str, int]]]:
    """Converts a stacked dataframe into 'hypercube' representation.

    Args:
        stacked_data: the data to convert
        wl_key: the main wavelength key.

    Returns:
        data: the data to interpolate (output of to_hypercube)
        params: the parameter sample points (output of to_hypercube)
        strings: non-interpolatable parameters (like strings)
            with their values (output of to_hypercube)

    Returns:
        the interpolated S-matrix at the values of **kwargs
    """
    if wl_key not in ["f", "wl"]:
        msg = f"Unsupported wl_key. Valid choices: 'wl', 'f'. Got: {wl_key}."
        raise ValueError(msg)
    df = stacked_data.copy()
    df["f"] = c / df["wl"].to_numpy()
    value_columns = [c for c in ["amp", "phi"] if c in df.columns]
    param_columns = [c for c in df.columns if c not in [*value_columns, "wl", "f"]]
    param_columns = [c for c in [*param_columns, wl_key] if c in df.columns]
    value_columns = ["amp", "phi"] if len(value_columns) > 2 else value_columns
    df = _sort_rows(
        cast(pd.DataFrame, df[[*param_columns, *value_columns]]),
        not_by=tuple(value_columns),
        wl_key=wl_key,
    )
    params: dict[str, Array | NDArray] = {
        c: np.asarray(df[c].unique()) for c in param_columns
    }
    data = df[value_columns].to_numpy()
    data = data.reshape(*(v.shape[0] for v in params.values()), len(value_columns))
    data = jnp.asarray(data)
    params["targets"] = np.array(value_columns, dtype=object)
    float_params, string_params = _extract_strings(params)
    return data, float_params, string_params


def _sort_rows(
    df: pd.DataFrame, not_by: tuple[str, ...] = ("amp", "phi"), wl_key: str = "wl"
) -> pd.DataFrame:
    by = [c for c in df.columns if c not in ["wl", "f", *not_by]] + [wl_key]
    return df.sort_values(by=by).reset_index(drop=True)


def _extract_strings(
    params: dict[str, Array | NDArray],
) -> tuple[dict[str, Array], dict[str, dict[str, int]]]:
    string_map = {}  # jax does not like strings
    new_params = {}
    for k, v in list(params.items()):
        if v.dtype == object:  # probably string array!
            string_map[k] = {s: i for i, s in enumerate(v)}
            new_params[k] = np.array(list(string_map[k].values()), dtype=int)
        else:
            new_params[k] = np.array(v, dtype=float)
    return new_params, string_map



import os
import pytest
import shutil
from papermill.engines import papermill_engines
from papermill.execute import raise_for_execution_errors
from papermill.iorw import load_notebook_node

TEST_DIR = os.path.dirname(os.path.abspath(__file__))
NBS_DIR = os.path.join(TEST_DIR, "nbs")
NBS_FAIL_DIR = os.path.join(NBS_DIR, "failed")


def get_kernel():
    kernel = os.environ.get("CONDA_DEFAULT_ENV", "base")
    if kernel == "base":
        kernel = "python3"
    return kernel


shutil.rmtree(NBS_FAIL_DIR, ignore_errors=True)
os.mkdir(NBS_FAIL_DIR)


def _find_notebooks(*dir):
    dir = os.path.abspath(os.path.join(TEST_DIR, *dir))
    for root, _, files in os.walk(dir):
        for file in files:
            if ("checkpoint" in file) or (not file.endswith(".ipynb")):
                continue
            yield os.path.join(root, file)


@pytest.mark.parametrize("path", sorted(_find_notebooks("nbs")))
def test_nbs(path):
    fn = os.path.basename(path)
    nb = load_notebook_node(path)
    nb = papermill_engines.execute_notebook_with_engine(
        engine_name=None,
        nb=nb,
        kernel_name=get_kernel(),
        input_path=path,
        output_path=None,
    )
    raise_for_execution_errors(nb, os.path.join(NBS_FAIL_DIR, fn))



""" SAX forward_only Backend """

from __future__ import annotations
from typing import Any, Dict

import jax.numpy as jnp
import networkx as nx

from ..netlist import Component
from ..saxtypes import Model, SCoo, scoo, SDict, sdict


def analyze_instances_forward(
    instances: Dict[str, Component],
    models: Dict[str, Model],
) -> Dict[str, SCoo]:
    instances, instances_old = {}, instances
    for k, v in instances_old.items():
        if not isinstance(v, Component):
            v = Component(**v)
        instances[k] = v
    model_names = set()
    for i in instances.values():
        model_names.add(i.component)
    dummy_models = {k: scoo(models[k]()) for k in model_names}
    dummy_instances = {}
    for k, i in instances.items():
        dummy_instances[k] = dummy_models[i.component]
    return dummy_instances


def analyze_circuit_forward(
    analyzed_instances: Dict[str, SDict],
    connections: Dict[str, str],
    ports: Dict[str, str],
) -> Any:
    return connections, ports


# import matplotlib.pyplot as plt
def evaluate_circuit_forward(
    analyzed: Any,
    instances: Dict[str, SDict],
) -> SDict:
    """Evaluate a circuit for the given sdicts using simple matrix multiplication."""
    connections, ports = analyzed
    edges = _graph_edges_directed(instances, connections, ports)

    graph = nx.DiGraph()
    graph.add_edges_from(edges)

    # Dictionary to store signals at each node
    circuit_sdict = {}
    for in_port in ports.keys():
        if in_port.startswith("in"):
            node_signals = {("", in_port): 1}
            bfs_output = nx.bfs_layers(graph, ("", in_port))
            for layer in bfs_output:
                layer_signals = {}
                for node in layer:
                    if node in node_signals:
                        signal = node_signals[node]
                        for neighbor in graph.successors(node):
                            transmission = graph[node][neighbor]["transmission"]
                            if neighbor in layer_signals:
                                layer_signals[neighbor] += signal * transmission
                            else:
                                layer_signals[neighbor] = signal * transmission
                node_signals.update(layer_signals)
            sdict = {
                (in_port, p2): v
                for (p1, p2), v in node_signals.items()
                if p1 == "" and p2.startswith("out")
            }
            circuit_sdict.update(sdict)
    return circuit_sdict


def _split_port(port: str) -> Tuple[str, str]:
    try:
        instance, port = port.split(",")
    except ValueError:
        (port,) = port.split(",")
        instance = ""
    return instance, port


def _graph_edges_directed(
    instances: Dict[str, SDict],
    connections: Dict[str, str],
    ports: Dict[str, str],
):
    one = jnp.array([1.0], dtype=float)
    edges_dict = {}
    edges_dict.update({_split_port(k): _split_port(v) for k, v in connections.items()})
    edges_dict.update({_split_port(k): _split_port(v) for k, v in ports.items()})
    edges = []
    for n1, n2 in edges_dict.items():
        if n1[0] == "" and n1[1].startswith("out"):
            edges += [(n2, n1, {"transmission": one})]
        else:
            edges += [(n1, n2, {"transmission": one})]

    for instance in instances:
        s = instances[instance]
        for (p1, p2), w in sdict(s).items():
            if p1.startswith("in") and p2.startswith("out"):
                edges += [
                    (
                        (instance, p1),
                        (instance, p2),
                        {"transmission": jnp.asarray(w, dtype=complex).ravel()},
                    )
                ]
    return edges



""" SAX General Utilities """

from __future__ import annotations

import inspect
import re
import warnings
from functools import lru_cache, partial, wraps
from hashlib import md5
from typing import (
    Any,
    Callable,
    Dict,
    Iterable,
    Iterator,
    NamedTuple,
    Tuple,
    Union,
    cast,
    overload,
)

import jax
import jax.numpy as jnp
import jax.scipy as jsp
import jax.scipy.linalg
import numpy as np
import orjson
from natsort import natsorted
from numpy.exceptions import ComplexWarning

from .saxtypes import (
    ComplexArrayND,
    FloatArrayND,
    Model,
    ModelFactory,
    SCoo,
    SDense,
    SDict,
    Settings,
    SType,
    is_mixedmode,
    is_model,
    is_model_factory,
    is_scoo,
    is_sdense,
    is_sdict,
)


def block_diag(*arrs: ComplexArrayND) -> ComplexArrayND:
    """create block diagonal matrix with arbitrary batch dimensions"""
    batch_shape = arrs[0].shape[:-2]

    N = 0
    for arr in arrs:
        if batch_shape != arr.shape[:-2]:
            raise ValueError("batch dimensions for given arrays don't match.")
        m, n = arr.shape[-2:]
        if m != n:
            raise ValueError("given arrays are not square.")
        N += n

    arrs = tuple(arr.reshape(-1, arr.shape[-2], arr.shape[-1]) for arr in arrs)
    batch_block_diag = jax.vmap(jsp.linalg.block_diag, in_axes=0, out_axes=0)
    block_diag = batch_block_diag(*arrs)
    return block_diag.reshape(*batch_shape, N, N)


def clean_string(s: str, dot="p", minus="m", other="_") -> str:
    """clean a string such that it is a valid python identifier"""
    s = s.strip()
    s = s.replace(".", dot)  # dot
    s = s.replace("-", minus)  # minus
    s = re.sub("[^0-9a-zA-Z_]", other, s)
    if s[0] in "0123456789":
        s = "_" + s
    if not s.isidentifier():
        raise ValueError(f"failed to clean string to a valid python identifier: {s}")
    return s


def copy_settings(settings: Settings) -> Settings:
    """copy a parameter dictionary"""
    return validate_settings(settings)  # validation also copies


def validate_settings(settings: Settings) -> Settings:
    """Validate a parameter dictionary"""
    _settings = {}
    for k, v in settings.items():
        if isinstance(v, dict):
            _settings[k] = validate_settings(v)
        else:
            _settings[k] = try_complex_float(v)
    return _settings


def try_complex_float(f: Any) -> Any:
    """try converting an object to float, return unchanged object on fail"""
    with warnings.catch_warnings():
        warnings.filterwarnings(action="error", category=ComplexWarning)
        try:
            return jnp.asarray(f, dtype=float)
        except ComplexWarning:
            return jnp.asarray(f, dtype=complex)
        except (ValueError, TypeError):
            return f
        finally:
            return f


def flatten_dict(dic: Dict[str, Any], sep: str = ",") -> Dict[str, Any]:
    """flatten a nested dictionary"""
    return _flatten_dict(dic, sep=sep)


def _flatten_dict(
    dic: Dict[str, Any], sep: str = ",", frozen: bool = False, parent_key: str = ""
) -> Dict[str, Any]:
    items = []
    for k, v in dic.items():
        new_key = parent_key + sep + k if parent_key else k
        if isinstance(v, dict):
            items.extend(
                _flatten_dict(v, sep=sep, frozen=frozen, parent_key=new_key).items()
            )
        else:
            items.append((new_key, v))

    return dict(items)


def unflatten_dict(dic, sep=","):
    """unflatten a flattened dictionary"""

    # from: https://gist.github.com/fmder/494aaa2dd6f8c428cede
    items = dict()

    for k, v in dic.items():
        keys = k.split(sep)
        sub_items = items
        for ki in keys[:-1]:
            if ki in sub_items:
                sub_items = sub_items[ki]
            else:
                sub_items[ki] = dict()
                sub_items = sub_items[ki]

        sub_items[keys[-1]] = v

    return items


def get_ports(S: Union[Model, SType]) -> Tuple[str, ...]:
    """get port names of a model or an stype

    .. note ::

        if a `Model` function is given in stead of an `SDict`, the function will be traced by JAX to obtain the port combinations of the resulting `SType`. Although this tracing of the function is 'cheap' in comparison to evaluating the model/circuit. It is not for free!  Use this function sparingly on your large `Model` or `circuit`!
    """
    if is_model(S):
        return _get_ports_from_model(cast(Model, S))
    elif is_sdict(S):
        S = cast(SDict, S)
        ports_set = {p1 for p1, _ in S} | {p2 for _, p2 in S}
        return tuple(natsorted(ports_set))
    elif is_scoo(S) or is_sdense(S):
        S = cast(SDense, S)
        *_, ports_map = S
        return tuple(natsorted(ports_map.keys()))
    else:
        raise ValueError("Could not extract ports for given S")


@lru_cache(maxsize=4096)  # cache to prevent future tracing
def _get_ports_from_model(model: Model) -> Tuple[str, ...]:
    # S: SType = jax.eval_shape(model)
    return get_ports(model())  # FIXME: this might be slow!


def get_port_combinations(S: Union[Model, SType]) -> Tuple[Tuple[str, str], ...]:
    """get port combinations of a model or an stype"""

    if is_model(S):
        S = cast(Model, S)
        return _get_port_combinations_from_model(S)
    elif is_sdict(S):
        S = cast(SDict, S)
        return tuple(S.keys())
    elif is_scoo(S):
        Si, Sj, _, pm = cast(SCoo, S)
        rpm = {int(i): str(p) for p, i in pm.items()}
        return tuple(natsorted((rpm[int(i)], rpm[int(j)]) for i, j in zip(Si, Sj)))
    elif is_sdense(S):
        _, pm = cast(SDense, S)
        return tuple(natsorted((p1, p2) for p1 in pm for p2 in pm))
    else:
        raise ValueError("Could not extract ports for given S")


@lru_cache(maxsize=4096)  # cache to prevent future tracing
def _get_port_combinations_from_model(model: Model) -> Tuple[Tuple[str, str], ...]:
    S: SType = jax.eval_shape(model)
    return get_port_combinations(S)


def get_settings(model: Union[Model, ModelFactory]) -> Settings:
    """Get the parameters of a SAX model function"""

    signature = inspect.signature(model)

    settings: Settings = {
        k: (v.default if not isinstance(v, dict) else v)
        for k, v in signature.parameters.items()
        if v.default is not inspect.Parameter.empty
    }

    # make sure an inplace operation of resulting dict does not change the
    # circuit parameters themselves
    return copy_settings(settings)


def grouped_interp(
    wl: FloatArrayND, wls: FloatArrayND, phis: FloatArrayND
) -> FloatArrayND:
    """Grouped phase interpolation

    .. note ::

        Grouped interpolation is useful to interpolate phase values where each datapoint
        is doubled (very close together) to give an indication of the phase
        variation at that point.

    .. warning ::

        this interpolation is only accurate in the range
        `[wls[0], wls[-2])` (`wls[-2]` not included). Any extrapolation
        outside these bounds can yield unexpected results!

    .. plot ::

        import sax
        import numpy as np
        import matplotlib.pyplot as plt
        wls = np.array([2.19999, 2.20001, 2.22499, 2.22501, 2.24999, 2.25001, 2.27499, 2.27501, 2.29999, 2.30001, 2.32499, 2.32501, 2.34999, 2.35001, 2.37499, 2.37501, 2.39999, 2.40001, 2.42499, 2.42501, 2.44999, 2.45001])
        phis = np.array([5.17317336, 5.1219654, 4.71259842, 4.66252492, 5.65699608, 5.60817922, 2.03697377, 1.98936119, 6.010146, 5.96358061, 4.96336733, 4.91777933, 5.13912198, 5.09451137, 0.22347545, 0.17979684, 2.74501894, 2.70224092, 0.10403192, 0.06214664, 4.83328794, 4.79225525])
        wl = np.linspace(wls.min(), wls.max(), 10000)
        phi = np.array(sax.grouped_interp(wl, wls, phis))

        _, ax = plt.subplots(2, 1, sharex=True, figsize=(14, 6))
        plt.sca(ax[0])
        plt.plot(1e3*wls, np.arange(wls.shape[0]), marker="o", ls="none")
        plt.grid(True)
        plt.ylabel("index")
        plt.sca(ax[1])
        plt.grid(True)
        plt.plot(1e3*wls, phis, marker="o", c="C1")
        plt.plot(1e3*wl, phi, c="C2")
        plt.xlabel("λ [nm]")
        plt.ylabel("φ")
        plt.show()

    """

    @partial(jax.vmap, in_axes=(0, None, None), out_axes=0)
    @jax.jit
    def _grouped_interp(
        wl: float,  # 0D array (not-vmapped) ; 1D array (vmapped)
        wls: FloatArrayND,  # 1D array
        phis: FloatArrayND,  # 1D array
    ) -> FloatArrayND:
        dphi_dwl = (phis[1::2] - phis[::2]) / (wls[1::2] - wls[::2])
        phis = phis[::2]
        wls = wls[::2]
        dwl = (wls[1:] - wls[:-1]).mean(0, keepdims=True)

        t = (wl - wls + 1e-5 * dwl) / dwl  # small offset to ensure no values are zero
        t = jnp.where(jnp.abs(t) < 1, t, 0)
        m0 = jnp.where(t > 0, size=1)[0]
        m1 = jnp.where(t < 0, size=1)[0]
        t = t[m0]
        wl0 = wls[m0]
        wl1 = wls[m1]
        phi0 = phis[m0]
        phi1 = phis[m1]
        dphi_dwl0 = dphi_dwl[m0]
        dphi_dwl1 = dphi_dwl[m1]
        _phi0 = phi0 - 0.5 * (wl1 - wl0) * (
            dphi_dwl0 * (t**2 - 2 * t) - dphi_dwl1 * t**2
        )
        _phi1 = phi1 - 0.5 * (wl1 - wl0) * (
            dphi_dwl0 * (t - 1) ** 2 - dphi_dwl1 * (t**2 - 1)
        )
        phis = jnp.arctan2(
            (1 - t) * jnp.sin(_phi0) + t * jnp.sin(_phi1),
            (1 - t) * jnp.cos(_phi0) + t * jnp.cos(_phi1),
        )
        return phis

    wl = jnp.asarray(wl)
    wls = jnp.asarray(wls)
    phis = jnp.asarray(phis) % (2 * jnp.pi)
    phis = jnp.where(phis > jnp.pi, phis - 2 * jnp.pi, phis)
    if not wls.ndim == 1:
        raise ValueError("grouped_interp: wls should be a 1D array")
    if not phis.ndim == 1:
        raise ValueError("grouped_interp: wls should be a 1D array")
    if not wls.shape == phis.shape:
        raise ValueError("grouped_interp: wls and phis shape does not match")
    return _grouped_interp(wl.reshape(-1), wls, phis).reshape(*wl.shape)


def merge_dicts(*dicts: Dict) -> Dict:
    """merge (possibly deeply nested) dictionaries"""
    if len(dicts) == 1:
        return dict(_generate_merged_dict(dicts[0], {}))
    elif len(dicts) == 2:
        return dict(_generate_merged_dict(dicts[0], dicts[1]))
    else:
        return merge_dicts(dicts[0], merge_dicts(*dicts[1:]))


def _generate_merged_dict(dict1: Dict, dict2: Dict) -> Iterator[Tuple[Any, Any]]:
    # inspired by https://stackoverflow.com/questions/7204805/how-to-merge-dictionaries-of-dictionaries
    keys = {
        **{k: None for k in dict1},
        **{k: None for k in dict2},
    }  # keep key order, values irrelevant
    for k in keys:
        if k in dict1 and k in dict2:
            v1, v2 = dict1[k], dict2[k]
            if isinstance(v1, dict) and isinstance(v2, dict):
                v = dict(_generate_merged_dict(v1, v2))
            else:
                # If one of the values is not a dict, you can't continue merging it.
                # Value from second dict overrides one in first and we move on.
                v = v2
        elif k in dict1:
            v = dict1[k]
        else:  # k in dict2:
            v = dict2[k]

        if isinstance(v, dict):
            yield (k, {**v})  # shallow copy of dict
        else:
            yield (k, v)


def mode_combinations(
    modes: Iterable[str], cross: bool = False
) -> Tuple[Tuple[str, str], ...]:
    """create mode combinations for a collection of given modes"""
    if cross:
        mode_combinations = natsorted((m1, m2) for m1 in modes for m2 in modes)
    else:
        mode_combinations = natsorted((m, m) for m in modes)
    return tuple(mode_combinations)


def reciprocal(sdict: SDict) -> SDict:
    """Make an SDict reciprocal"""
    if is_sdict(sdict):
        return {
            **{(p1, p2): v for (p1, p2), v in sdict.items()},
            **{(p2, p1): v for (p1, p2), v in sdict.items()},
        }
    else:
        raise ValueError("sax.reciprocal is only valid for SDict types")


@overload
def rename_params(model: ModelFactory, renamings: Dict[str, str]) -> ModelFactory:
    ...


@overload
def rename_params(model: Model, renamings: Dict[str, str]) -> Model:
    ...


def rename_params(
    model: Union[Model, ModelFactory], renamings: Dict[str, str]
) -> Union[Model, ModelFactory]:
    """rename the parameters of a `Model` or `ModelFactory` given
    a renamings mapping old parameter names to new."""

    reversed_renamings = {v: k for k, v in renamings.items()}
    if len(reversed_renamings) < len(renamings):
        raise ValueError("Multiple old names point to the same new name!")

    if is_model_factory(model):
        old_model_factory = cast(ModelFactory, model)
        old_settings = get_settings(model)

        @wraps(old_model_factory)
        def new_model_factory(**settings):
            old_settings = {
                reversed_renamings.get(k, k): v for k, v in settings.items()
            }
            model = old_model_factory(**old_settings)
            return rename_params(model, renamings)

        new_settings = {renamings.get(k, k): v for k, v in old_settings.items()}
        _replace_kwargs(new_model_factory, **new_settings)

        return new_model_factory

    elif is_model(model):
        old_model = cast(Model, model)
        old_settings = get_settings(model)

        @wraps(old_model)
        def new_model(**settings):
            old_settings = {
                reversed_renamings.get(k, k): v for k, v in settings.items()
            }
            return old_model(**old_settings)

        new_settings = {renamings.get(k, k): v for k, v in old_settings.items()}
        _replace_kwargs(new_model, **new_settings)

        return new_model

    else:
        raise ValueError(
            "rename_params should be used to decorate a Model or ModelFactory."
        )


def _replace_kwargs(func: Callable, **kwargs: Any):
    """Change the kwargs signature of a function"""
    sig = inspect.signature(func)
    settings = [
        inspect.Parameter(k, inspect.Parameter.KEYWORD_ONLY, default=v)
        for k, v in kwargs.items()
    ]
    func.__signature__ = sig.replace(parameters=settings)


@overload
def rename_ports(S: SDict, renamings: Dict[str, str]) -> SDict:
    ...


@overload
def rename_ports(S: SCoo, renamings: Dict[str, str]) -> SCoo:
    ...


@overload
def rename_ports(S: SDense, renamings: Dict[str, str]) -> SDense:
    ...


@overload
def rename_ports(S: Model, renamings: Dict[str, str]) -> Model:
    ...


@overload
def rename_ports(S: ModelFactory, renamings: Dict[str, str]) -> ModelFactory:
    ...


def rename_ports(
    S: Union[SType, Model, ModelFactory], renamings: Dict[str, str]
) -> Union[SType, Model, ModelFactory]:
    """rename the ports of an `SDict`, `Model` or `ModelFactory` given
    a renamings mapping old port names to new."""
    if is_scoo(S):
        Si, Sj, Sx, ports_map = cast(SCoo, S)
        ports_map = {renamings[p]: i for p, i in ports_map.items()}
        return Si, Sj, Sx, ports_map
    elif is_sdense(S):
        Sx, ports_map = cast(SDense, S)
        ports_map = {renamings[p]: i for p, i in ports_map.items()}
        return Sx, ports_map
    elif is_sdict(S):
        sdict = cast(SDict, S)
        original_ports = get_ports(sdict)
        assert len(renamings) == len(original_ports)
        return {(renamings[p1], renamings[p2]): v for (p1, p2), v in sdict.items()}
    elif is_model(S):
        old_model = cast(Model, S)

        @wraps(old_model)
        def new_model(**settings) -> SType:
            return rename_ports(old_model(**settings), renamings)

        return new_model
    elif is_model_factory(S):
        old_model_factory = cast(ModelFactory, S)

        @wraps(old_model_factory)
        def new_model_factory(**settings) -> Callable[..., SType]:
            return rename_ports(old_model_factory(**settings), renamings)

        return new_model_factory
    else:
        raise ValueError("Cannot rename ports for type {type(S)}")


def update_settings(settings: Settings, *compnames: str, **kwargs: Any) -> Settings:
    """update a nested settings dictionary

    .. note ::

        1. Even though it's possible to update parameter dictionaries in place,
        this function is convenient to apply certain parameters (e.g. wavelength
        'wl' or temperature 'T') globally.
        2. This operation never updates the given settings dictionary inplace.
        3. Any non-float keyword arguments will be silently ignored.

    """
    _settings = {}
    if not compnames:
        for k, v in settings.items():
            if isinstance(v, dict):
                _settings[k] = update_settings(v, **kwargs)
            else:
                if k in kwargs:
                    _settings[k] = try_complex_float(kwargs[k])
                else:
                    _settings[k] = try_complex_float(v)
    else:
        for k, v in settings.items():
            if isinstance(v, dict):
                if k == compnames[0]:
                    _settings[k] = update_settings(v, *compnames[1:], **kwargs)
                else:
                    _settings[k] = v
            else:
                _settings[k] = try_complex_float(v)
    return _settings


def validate_not_mixedmode(S: SType):
    """validate that an stype is not 'mixed mode' (i.e. invalid)

    Args:
        S: the stype to validate
    """

    if is_mixedmode(S):  # mixed mode
        raise ValueError(
            "Given SType is neither multimode or singlemode. Please check the port "
            "names: they should either ALL contain the '@' separator (multimode) "
            "or NONE should contain the '@' separator (singlemode)."
        )


def validate_multimode(S: SType, modes=("te", "tm")) -> None:
    """validate that an stype is multimode and that the given modes are present."""
    try:
        current_modes = set(p.split("@")[1] for p in get_ports(S))
    except IndexError:
        raise ValueError("The given stype is not multimode.")
    for mode in modes:
        if mode not in current_modes:
            raise ValueError(
                f"Could not find mode '{mode}' in one of the multimode models."
            )


def validate_sdict(sdict: Any) -> None:
    """Validate an `SDict`"""

    if not isinstance(sdict, dict):
        raise ValueError("An SDict should be a dictionary.")
    for ports in sdict:
        if not isinstance(ports, tuple) and not len(ports) == 2:
            raise ValueError(f"SDict keys should be length-2 tuples. Got {ports}")
        p1, p2 = ports
        if not isinstance(p1, str) or not isinstance(p2, str):
            raise ValueError(
                f"SDict ports should be strings. Got {ports} "
                f"({type(ports[0])}, {type(ports[1])})"
            )


def get_inputs_outputs(ports: Tuple[str, ...]):
    inputs = tuple(p for p in ports if p.lower().startswith("in"))
    outputs = tuple(p for p in ports if not p.lower().startswith("in"))
    if not inputs:
        inputs = tuple(p for p in ports if not p.lower().startswith("out"))
        outputs = tuple(p for p in ports if p.lower().startswith("out"))
    return inputs, outputs


def hash_dict(dic: Dict) -> int:
    return int(
        md5(
            orjson.dumps(
                _numpyfy(dic), option=orjson.OPT_SERIALIZE_NUMPY | orjson.OPT_SORT_KEYS
            )
        ).hexdigest(),
        16,
    )


def _numpyfy(obj: Any):
    if not isinstance(obj, dict):
        return np.asarray(obj)
    else:
        return {k: _numpyfy(v) for k, v in obj.items()}


class Normalization(NamedTuple):
    mean: ComplexArrayND
    std: ComplexArrayND


def normalization(x: ComplexArrayND, axis=None):
    if axis is None:
        return Normalization(x.mean(), x.std())
    else:
        return Normalization(x.mean(axis), x.std(axis))


def cartesian_product(*arrays: ComplexArrayND) -> ComplexArrayND:
    """calculate the n-dimensional cartesian product of an arbitrary number of arrays"""
    ixarrays = jnp.ix_(*arrays)
    barrays = jnp.broadcast_arrays(*ixarrays)
    sarrays = jnp.stack(barrays, -1)
    assert isinstance(sarrays, jnp.ndarray)
    product = sarrays.reshape(-1, sarrays.shape[-1])
    assert isinstance(product, jnp.ndarray)
    return product


def normalize(x: ComplexArrayND, normalization: Normalization) -> Tuple[ComplexArrayND]:
    """normalize an array with a given mean and standard deviation"""
    mean, std = normalization
    return (x - mean) / std


def denormalize(x: ComplexArrayND, normalization: Normalization) -> ComplexArrayND:
    """denormalize an array with a given mean and standard deviation"""
    mean, std = normalization
    return x * std + mean



from functools import partial

import sax

SAMPLE_NETLIST = {
    "instances": {
        "lft": "coupler",
        "top": "waveguide",
        "btm": "waveguide",
        "rgt": "coupler",
    },
    "connections": {
        "lft,out0": "btm,in0",
        "btm,out0": "rgt,in0",
        "lft,out1": "top,in0",
        "top,out0": "rgt,in1",
    },
    "ports": {
        "in0": "lft,in0",
        "in1": "lft,in1",
        "out0": "rgt,out0",
        "out1": "rgt,out1",
    },
}

SAMPLE_NETLIST2 = {
    "instances": {
        "lft": sax.models.coupler,
        "top": partial(sax.models.straight, length=10.0),
        "btm": partial(sax.models.straight, length=30.0),
        "rgt": sax.models.coupler,
    },
    "connections": {
        "lft,out0": "btm,in0",
        "btm,out0": "rgt,in0",
        "lft,out1": "top,in0",
        "top,out0": "rgt,in1",
    },
    "ports": {
        "in0": "lft,in0",
        "in1": "lft,in1",
        "out0": "rgt,out0",
        "out1": "rgt,out1",
    },
}


def test_empty_netlist():
    net = sax.Netlist()
    assert isinstance(net, sax.Netlist)


def test_coercing_instances():
    net = sax.Netlist.model_validate({"instances": {"mmi": "mmi"}})
    assert isinstance(net, sax.Netlist)


def test_netlist():
    recnet = sax.Netlist.model_validate(SAMPLE_NETLIST)
    assert isinstance(recnet, sax.Netlist)


def test_recursive_netlist():
    recnet = sax.RecursiveNetlist.model_validate({"top_level": SAMPLE_NETLIST})
    assert isinstance(recnet, sax.RecursiveNetlist)


def test_netlist_function_from_dict():
    net = sax.netlist(SAMPLE_NETLIST)
    assert isinstance(net, sax.RecursiveNetlist)
    assert "top_level" in net.root


def test_netlist_function_from_netlist():
    net = sax.netlist(sax.Netlist.model_validate(SAMPLE_NETLIST))
    assert isinstance(net, sax.RecursiveNetlist)
    assert "top_level" in net.root


def test_netlist_function_from_recursive_netlist():
    net = sax.netlist(
        sax.RecursiveNetlist.model_validate({"top_level": SAMPLE_NETLIST})
    )
    assert isinstance(net, sax.RecursiveNetlist)
    assert "top_level" in net.root


def test_netlist_using_instance_functions():
    net = sax.netlist(SAMPLE_NETLIST2)
    print(net)
    assert isinstance(net, sax.RecursiveNetlist)


if __name__ == "__main__":
    test_netlist_using_instance_functions()



""" SAX Types and type coercions """

from __future__ import annotations

import functools
import inspect
from collections.abc import Callable as CallableABC
from typing import Any, Callable, Dict, Tuple, Union, cast, overload

import jax.numpy as jnp
import numpy as np
from jaxtyping import Array as Array
from jaxtyping import ArrayLike as ArrayLike
from jaxtyping import Complex as Complex
from jaxtyping import Float as Float
from jaxtyping import Int as Int
from natsort import natsorted

IntArray1D = Int[Array, " dim"]
""" One dimensional integer array """

FloatArray1D = Complex[Array, " dim"]
""" One dimensional float array """

ComplexArray1D = Complex[Array, " dim"]
""" One dimensional complex array """

IntArrayND = Int[Array, "..."]
""" N-dimensional integer array """

FloatArrayND = Complex[Array, "..."]
""" N-dimensional float array """

ComplexArrayND = Complex[Array, "..."]
""" N-dimensional complex array """

PortMap = Dict[str, int]
""" A mapping from a port name (str) to a port index (int) """

PortCombination = Tuple[str, str]
""" A combination of two port names (str, str) """

SDict = Dict[PortCombination, ComplexArrayND]
""" A mapping from a port combination to an S-parameter or an array of S-parameters

Example:

.. code-block::

    sdict: sax.SDict = {
        ("in0", "out0"): 3.0,
    }

"""

SDense = Tuple[ComplexArrayND, PortMap]
""" A dense S-matrix (2D array) or multidimensional batched S-matrix (N+2)-D array
combined with a port map. If (N+2)-D array the S-matrix dimensions are the last two.

Example:

.. code-block::

    Sd = jnp.arange(9, dtype=float).reshape(3, 3)
    port_map = {"in0": 0, "in1": 2, "out0": 1}
    sdense = Sd, port_map

"""

SCoo = Tuple[IntArray1D, IntArray1D, ComplexArrayND, PortMap]
""" A sparse S-matrix in COO format (recommended for internal library use only)

An `SCoo` is a sparse matrix based representation of an S-matrix consisting of three arrays and a port map.
The three arrays represent the input port indices [`int`], output port indices [`int`] and the S-matrix values [`ComplexFloat`] of the sparse matrix.
The port map maps a port name [`str`] to a port index [`int`]. Only these four arrays **together** and in this specific **order** are considered a valid `SCoo` representation!

Example:

.. code-block::

    Si = jnp.arange(3, dtype=int)
    Sj = jnp.array([0, 1, 0], dtype=int)
    Sx = jnp.array([3.0, 4.0, 1.0])
    port_map = {"in0": 0, "in1": 2, "out0": 1}
    scoo: sax.SCoo = (Si, Sj, Sx, port_map)

"""

Settings = Dict[str, Union["Settings", FloatArrayND, ComplexArrayND]]
""" A (possibly recursive) mapping from a setting name to a float or complex value or array

Example:

.. code-block::

    mzi_settings: sax.Settings = {
        "wl": 1.5,  # global settings
        "lft": {"coupling": 0.5},  # settings for the left coupler
        "top": {"neff": 3.4},  # settings for the top waveguide
        "rgt": {"coupling": 0.3},  # settings for the right coupler
    }

"""

SType = Union[SDict, SCoo, SDense]
""" An SDict, SDense or SCOO """

Model = Callable[..., SType]
""" A keyword-only function producing an SType """

ModelFactory = Callable[..., Model]
""" A keyword-only function producing a Model """


def is_float(x: Any) -> bool:
    """Check if an object is a `Float`"""
    if isinstance(x, float):
        return True
    if isinstance(x, np.ndarray):
        return x.dtype in (np.float16, np.float32, np.float64, np.float128)
    if isinstance(x, jnp.ndarray):
        return x.dtype in (jnp.float16, jnp.float32, jnp.float64)
    return False


def is_complex(x: Any) -> bool:
    """check if an object is a `ComplexFloat`"""
    if isinstance(x, complex):
        return True
    if isinstance(x, np.ndarray):
        return x.dtype in (np.complex64, np.complex128)
    if isinstance(x, jnp.ndarray):
        return x.dtype in (jnp.complex64, jnp.complex128)
    return False


def is_complex_float(x: Any) -> bool:
    """check if an object is either a `ComplexFloat` or a `Float`"""
    return is_float(x) or is_complex(x)


def is_sdict(x: Any) -> bool:
    """check if an object is an `SDict` (a SAX S-dictionary)"""
    return isinstance(x, dict)


def is_scoo(x: Any) -> bool:
    """check if an object is an `SCoo` (a SAX sparse S representation in COO-format)"""
    return isinstance(x, (tuple, list)) and len(x) == 4


def is_sdense(x: Any) -> bool:
    """check if an object is an `SDense` (a SAX dense S-matrix representation)"""
    return isinstance(x, (tuple, list)) and len(x) == 2


def is_model(model: Any) -> bool:
    """check if a callable is a `Model` (a callable returning an `SType`)"""
    if not callable(model):
        return False
    try:
        sig = inspect.signature(model)
    except ValueError:
        return False
    for param in sig.parameters.values():
        if param.default is inspect.Parameter.empty:
            return False  # a proper SAX model does not have any positional arguments.
    if _is_callable_annotation(sig.return_annotation):  # model factory
        return False
    return True


def _is_callable_annotation(annotation: Any) -> bool:
    """check if an annotation is `Callable`-like"""
    if isinstance(annotation, str):
        # happens when
        # was imported at the top of the file...
        return annotation.startswith("Callable") or annotation.endswith("Model")
        # TODO: this is not a very robust check...
    try:
        return annotation.__origin__ == CallableABC
    except AttributeError:
        return False


def is_model_factory(model: Any) -> bool:
    """check if a callable is a model function."""
    if not callable(model):
        return False
    sig = inspect.signature(model)
    if _is_callable_annotation(sig.return_annotation):  # model factory
        return True
    return False


def validate_model(model: Callable):
    """Validate the parameters of a model"""
    positional_arguments = []
    for param in inspect.signature(model).parameters.values():
        if param.default is inspect.Parameter.empty:
            positional_arguments.append(param.name)
    if positional_arguments:
        raise ValueError(
            f"model '{model}' takes positional "
            f"arguments {', '.join(positional_arguments)} "
            "and hence is not a valid SAX Model! "
            "A SAX model should ONLY take keyword arguments (or no arguments at all)."
        )


def is_stype(stype: Any) -> bool:
    """check if an object is an SDict, SCoo or SDense"""
    return is_sdict(stype) or is_scoo(stype) or is_sdense(stype)


def is_singlemode(S: Any) -> bool:
    """check if an stype is single mode"""
    if not is_stype(S):
        return False
    ports = _get_ports(S)
    return not any(("@" in p) for p in ports)


def _get_ports(S: SType):
    if is_sdict(S):
        S = cast(SDict, S)
        ports_set = {p1 for p1, _ in S} | {p2 for _, p2 in S}
        return tuple(natsorted(ports_set))
    else:
        *_, ports_map = S
        assert isinstance(ports_map, dict)
        return tuple(natsorted(ports_map.keys()))


def is_multimode(S: Any) -> bool:
    """check if an stype is single mode"""
    if not is_stype(S):
        return False

    ports = _get_ports(S)
    return all(("@" in p) for p in ports)


def is_mixedmode(S: Any) -> bool:
    """check if an stype is neither single mode nor multimode (hence invalid)"""
    return not is_singlemode(S) and not is_multimode(S)


@overload
def sdict(S: Model) -> Model:
    ...


@overload
def sdict(S: SType) -> SDict:
    ...


def sdict(S: Union[Model, SType]) -> Union[Model, SType]:
    """Convert an `SCoo` or `SDense` to `SDict`"""

    if is_model(S):
        model = cast(Model, S)

        @functools.wraps(model)
        def wrapper(**kwargs):
            return sdict(model(**kwargs))

        return wrapper

    elif is_scoo(S):
        x_dict = _scoo_to_sdict(*cast(SCoo, S))
    elif is_sdense(S):
        x_dict = _sdense_to_sdict(*cast(SDense, S))
    elif is_sdict(S):
        x_dict = cast(SDict, S)
    else:
        raise ValueError("Could not convert arguments to sdict.")

    return x_dict


def _scoo_to_sdict(
    Si: IntArray1D,
    Sj: IntArray1D,
    Sx: ComplexArrayND,
    ports_map: Dict[str, int],
) -> SDict:
    sdict = {}
    inverse_ports_map = {int(i): p for p, i in ports_map.items()}
    for i, (si, sj) in enumerate(zip(Si, Sj)):
        input_port = inverse_ports_map.get(int(si), "")
        output_port = inverse_ports_map.get(int(sj), "")
        sdict[input_port, output_port] = Sx[..., i]
    sdict = {(p1, p2): v for (p1, p2), v in sdict.items() if p1 and p2}
    return sdict


def _sdense_to_sdict(S: Array, ports_map: Dict[str, int]) -> SDict:
    sdict = {}
    for p1, i in ports_map.items():
        for p2, j in ports_map.items():
            sdict[p1, p2] = S[..., i, j]
    return sdict


@overload
def scoo(S: Callable) -> Callable:
    ...


@overload
def scoo(S: SType) -> SCoo:
    ...


def scoo(S: Union[Callable, SType]) -> Union[Callable, SCoo]:
    """Convert an `SDict` or `SDense` to `SCoo`"""

    if is_model(S):
        model = cast(Model, S)

        @functools.wraps(model)
        def wrapper(**kwargs):
            return scoo(model(**kwargs))

        return wrapper

    elif is_scoo(S):
        S = cast(SCoo, S)
    elif is_sdense(S):
        S = _sdense_to_scoo(*cast(SDense, S))
    elif is_sdict(S):
        S = _sdict_to_scoo(cast(SDict, S))
    else:
        raise ValueError("Could not convert arguments to scoo.")

    return S


def _consolidate_sdense(S, pm):
    idxs = list(pm.values())
    S = S[..., idxs, :][..., :, idxs]
    pm = {p: i for i, p in enumerate(pm)}
    return S, pm


def _sdense_to_scoo(S: Array, ports_map: Dict[str, int]) -> SCoo:
    S, ports_map = _consolidate_sdense(S, ports_map)
    Sj, Si = jnp.meshgrid(jnp.arange(S.shape[-1]), jnp.arange(S.shape[-2]))
    return Si.ravel(), Sj.ravel(), S.reshape(*S.shape[:-2], -1), ports_map


def _sdict_to_scoo(sdict: SDict) -> SCoo:
    all_ports = {}
    for p1, p2 in sdict:
        all_ports[p1] = None
        all_ports[p2] = None
    ports_map = {p: int(i) for i, p in enumerate(all_ports)}
    Sx = jnp.stack(jnp.broadcast_arrays(*sdict.values()), -1)
    Si = jnp.array([ports_map[p] for p, _ in sdict])
    Sj = jnp.array([ports_map[p] for _, p in sdict])
    return Si, Sj, Sx, ports_map


@overload
def sdense(S: Callable) -> Callable:
    ...


@overload
def sdense(S: SType) -> SDense:
    ...


def sdense(S: Union[Callable, SType]) -> Union[Callable, SDense]:
    """Convert an `SDict` or `SCoo` to `SDense`"""

    if is_model(S):
        model = cast(Model, S)

        @functools.wraps(model)
        def wrapper(**kwargs):
            return sdense(model(**kwargs))

        return wrapper

    if is_sdict(S):
        S = _sdict_to_sdense(cast(SDict, S))
    elif is_scoo(S):
        S = _scoo_to_sdense(*cast(SCoo, S))
    elif is_sdense(S):
        S = cast(SDense, S)
    else:
        raise ValueError("Could not convert arguments to sdense.")

    return S


def _scoo_to_sdense(
    Si: Array, Sj: Array, Sx: Array, ports_map: Dict[str, int]
) -> SDense:
    n_col = len(ports_map)
    S = jnp.zeros((*Sx.shape[:-1], n_col, n_col), dtype=complex)
    S = S.at[..., Si, Sj].add(Sx)
    return S, ports_map


def _sdict_to_sdense(sdict: SDict) -> SDense:
    Si, Sj, Sx, ports_map = _sdict_to_scoo(sdict)
    return _scoo_to_sdense(Si, Sj, Sx, ports_map)


def modelfactory(func):
    """Decorator that marks a function as `ModelFactory`"""
    sig = inspect.signature(func)
    if _is_callable_annotation(sig.return_annotation):  # already model factory
        return func
    func.__signature__ = sig.replace(return_annotation=Model)
    return func



""" SAX netlist parsing and utilities """

from __future__ import annotations

import os
import re
import warnings
from copy import deepcopy
from functools import lru_cache, partial
from typing import Any, Literal, TypedDict, overload

import black
import networkx as nx
import numpy as np
import yaml
from natsort import natsorted
from pydantic import AfterValidator
from pydantic import BaseModel as _BaseModel
from pydantic import (
    BeforeValidator,
    ConfigDict,
    Field,
    RootModel,
    ValidationError,
    model_validator,
)
from typing_extensions import Annotated

from .utils import clean_string, hash_dict


class NetlistDict(TypedDict):
    instances: dict
    connections: dict[str, str]
    ports: dict[str, str]
    settings: dict[str, Any]


RecursiveNetlistDict = dict[str, NetlistDict]


class BaseModel(_BaseModel):
    model_config = ConfigDict(
        extra="ignore",
        json_encoders={np.ndarray: lambda arr: np.round(arr, 12).tolist()},
    )

    def __repr__(self):
        s = super().__repr__()
        s = black.format_str(s, mode=black.Mode())
        return s

    def __str__(self):
        return self.__repr__()

    def __hash__(self):
        return hash_dict(self.model_dump())


def _validate_str(s: str, what="component"):
    if "," in s:
        raise ValueError(f"Invalid {what} string. Should not contain ','. Got: {s}")
    s = s.split("$")[0]
    s = clean_string(s)
    return s


ComponentStr = Annotated[str, AfterValidator(_validate_str)]


class Component(BaseModel):
    component: ComponentStr
    settings: dict[str, Any] = Field(default_factory=dict)


PortPlacement = Literal["ce", "cw", "nc", "ne", "nw", "sc", "se", "sw", "cc", "center"]


class Placement(BaseModel):
    x: str | float = 0.0
    y: str | float = 0.0
    dx: str | float = 0.0
    dy: str | float = 0.0
    rotation: float = 0.0
    mirror: bool = False
    xmin: str | float | None = None
    xmax: str | float | None = None
    ymin: str | float | None = None
    ymax: str | float | None = None
    port: str | PortPlacement | None = None


def _component_from_partial(p: partial):
    settings = {}
    f: Any = p
    while isinstance(f, partial):
        if f.args:
            raise ValueError(
                "SAX circuits and netlists don't support partials "
                "with positional arguments."
            )
        settings = {**f.keywords, **settings}
        f = f.func
    if not callable(f):
        raise ValueError("partial of non-callable.")
    return Component(component=f.__name__, settings=settings)


def _coerce_component(obj: Any) -> Component:
    if isinstance(obj, str):
        return Component(component=obj)
    elif isinstance(obj, partial):
        return _component_from_partial(obj)
    elif callable(obj):
        return _coerce_component(obj.__name__)
    elif isinstance(obj, dict) and "info" in obj:
        info = obj.pop("info", {})
        settings = obj.pop("settings", {})
        obj["settings"] = {**settings, **info}
    return Component.model_validate(obj)


CoercingComponent = Annotated[Component, BeforeValidator(_coerce_component)]


_validate_instance_str = partial(_validate_str, what="instance")
_validate_port_str = partial(_validate_str, what="port")


def _validate_instance_port_str(s: str):
    parts = s.split(",")
    if len(parts) != 2:
        raise ValueError(
            f"Invalid instance,port string. Should contain exactly one ','. Got: {s}"
        )
    i, p = parts
    i = _validate_instance_str(i)
    p = _validate_port_str(p)
    return f"{i},{p}"


InstanceStr = Annotated[str, AfterValidator(_validate_instance_str)]
PortStr = Annotated[str, AfterValidator(_validate_port_str)]
InstancePortStr = Annotated[str, AfterValidator(_validate_instance_port_str)]


def _nets_to_connections(nets: list[dict], connections: dict):
    connections = {k: v for k, v in connections.items()}
    inverse_connections = {v: k for k, v in connections.items()}

    def _is_connected(p):
        return (p in connections) or (p in inverse_connections)

    def _add_connection(p, q):
        connections[p] = q
        inverse_connections[q] = p

    def _get_connected_port(p):
        if p in connections:
            return connections[p]
        else:
            return inverse_connections[p]

    for net in nets:
        p = net["p1"]
        q = net["p2"]
        if _is_connected(p):
            _q = _get_connected_port(p)
            raise ValueError(
                "SAX currently does not support multiply connected ports. "
                f"Got {p}<->{q} and {p}<->{_q}"
            )
        if _is_connected(q):
            _p = _get_connected_port(q)
            raise ValueError(
                "SAX currently does not support multiply connected ports. "
                f"Got {p}<->{q} and {_p}<->{q}"
            )
        _add_connection(p, q)
    return connections


class Netlist(BaseModel):
    instances: dict[InstanceStr, CoercingComponent] = Field(default_factory=dict)
    connections: dict[InstancePortStr, InstancePortStr] = Field(default_factory=dict)
    ports: dict[PortStr, InstancePortStr] = Field(default_factory=dict)
    placements: dict[InstanceStr, Placement] = Field(default_factory=dict)
    settings: dict[str, Any] = Field(default_factory=dict)  # TODO: use this

    @model_validator(mode="before")
    @classmethod
    def coerce_nets_into_connections(cls, netlist: dict):
        if not isinstance(netlist, dict):
            return netlist
        if "nets" in netlist:
            nets = netlist.pop("nets", [])
            connections = netlist.pop("connections", {})
            connections = _nets_to_connections(nets, connections)
            netlist["connections"] = connections
        return netlist


class RecursiveNetlist(RootModel):
    root: dict[str, Netlist]

    model_config = ConfigDict(
        json_encoders={np.ndarray: lambda arr: np.round(arr, 12).tolist()},
    )

    def __repr__(self):
        s = super().__repr__()
        s = black.format_str(s, mode=black.Mode())
        return s

    def __str__(self):
        return self.__repr__()

    def __hash__(self):
        return hash_dict(self.model_dump())


AnyNetlist = Netlist | NetlistDict | RecursiveNetlist | RecursiveNetlistDict


def netlist(
    netlist: Any, with_unconnected_instances: bool = True, with_placements=True
) -> RecursiveNetlist:
    """return a netlist from a given dictionary"""
    if isinstance(netlist, RecursiveNetlist):
        net = netlist
    elif isinstance(netlist, Netlist):
        net = RecursiveNetlist(root={"top_level": netlist})
    elif isinstance(netlist, dict):
        if is_recursive(netlist):
            net = RecursiveNetlist.model_validate(netlist)
        else:
            flat_net = Netlist.model_validate(netlist)
            net = RecursiveNetlist.model_validate({"top_level": flat_net})
    else:
        raise ValueError(
            "Invalid argument for `netlist`. "
            "Expected type: dict | Netlist | RecursiveNetlist. "
            f"Got: {type(netlist)}."
        )
    if not with_unconnected_instances:
        recnet_dict: RecursiveNetlistDict = _remove_unused_instances(net.model_dump())
        net = RecursiveNetlist.model_validate(recnet_dict)
    if not with_placements:
        for _net in net.root.values():
            _net.placements = {}
    return net


def flatten_netlist(recnet: RecursiveNetlistDict, sep: str = "~"):
    first_name = list(recnet.keys())[0]
    net = _copy_netlist(recnet[first_name])
    _flatten_netlist(recnet, net, sep)
    return net


@lru_cache()
def load_netlist(pic_path: str) -> Netlist:
    with open(pic_path, "r") as file:
        net = yaml.safe_load(file.read())
    return Netlist.model_validate(net)


@lru_cache()
def load_recursive_netlist(pic_path: str, ext: str = ".yml"):
    folder_path = os.path.dirname(os.path.abspath(pic_path))

    def _clean_string(path: str) -> str:
        return clean_string(re.sub(ext, "", os.path.split(path)[-1]))

    # the circuit we're interested in should come first:
    netlists: dict[str, Netlist] = {_clean_string(pic_path): Netlist()}

    for filename in os.listdir(folder_path):
        path = os.path.join(folder_path, filename)
        if not os.path.isfile(path) or not path.endswith(ext):
            continue
        netlists[_clean_string(path)] = load_netlist(path)

    return RecursiveNetlist.model_validate(netlists)


def is_recursive(netlist: AnyNetlist):
    if isinstance(netlist, RecursiveNetlist):
        return True
    elif isinstance(netlist, dict):
        return "instances" not in netlist
    else:
        return False


def is_not_recursive(netlist: AnyNetlist):
    return not is_recursive(netlist)


def get_netlist_instances_by_prefix(
    recursive_netlist: RecursiveNetlist,
    prefix: str,
):
    """
    Returns a list of all instances with a given prefix in a recursive netlist.

    Args:
        recursive_netlist: The recursive netlist to search.
        prefix: The prefix to search for.

    Returns:
        A list of all instances with the given prefix.
    """
    recursive_netlist_root = recursive_netlist.model_dump()
    result = []
    for key in recursive_netlist_root.keys():
        if key.startswith(prefix):
            result.append(key)
    return result


def get_component_instances(
    recursive_netlist: RecursiveNetlist,
    top_level_prefix: str,
    component_name_prefix: str,
):
    """
    Returns a dictionary of all instances of a given component in a recursive netlist.

    Args:
        recursive_netlist: The recursive netlist to search.
        top_level_prefix: The prefix of the top level instance.
        component_name_prefix: The name of the component to search for.

    Returns:
        A dictionary of all instances of the given component.
    """
    instance_names = []
    recursive_netlist_root = recursive_netlist.model_dump()

    # Should only be one in a netlist-to-digraph. Can always be very specified.
    top_level_prefixes = get_netlist_instances_by_prefix(
        recursive_netlist, prefix=top_level_prefix
    )
    top_level_prefix = top_level_prefixes[0]
    for key in recursive_netlist_root[top_level_prefix]["instances"]:
        if recursive_netlist_root[top_level_prefix]["instances"][key][
            "component"
        ].startswith(component_name_prefix):
            # Note priority encoding on match.
            instance_names.append(key)
    return {component_name_prefix: instance_names}


def _remove_unused_instances(recursive_netlist: RecursiveNetlistDict):
    recursive_netlist = {**recursive_netlist}

    for name, flat_netlist in recursive_netlist.items():
        recursive_netlist[name] = _remove_unused_instances_flat(flat_netlist)

    return recursive_netlist


def _get_connectivity_netlist(netlist):
    connectivity_netlist = {
        "instances": natsorted(netlist["instances"]),
        "connections": [
            (c1.split(",")[0], c2.split(",")[0])
            for c1, c2 in netlist["connections"].items()
        ],
        "ports": [(p, c.split(",")[0]) for p, c in netlist["ports"].items()],
    }
    return connectivity_netlist


def _get_connectivity_graph(netlist):
    graph = nx.Graph()
    connectivity_netlist = _get_connectivity_netlist(netlist)
    for name in connectivity_netlist["instances"]:
        graph.add_node(name)
    for c1, c2 in connectivity_netlist["connections"]:
        graph.add_edge(c1, c2)
    for c1, c2 in connectivity_netlist["ports"]:
        graph.add_edge(c1, c2)
    return graph


def _get_nodes_to_remove(graph, netlist):
    nodes = set()
    for port in netlist["ports"]:
        nodes |= nx.descendants(graph, port)
    nodes_to_remove = set(graph.nodes) - nodes
    return list(nodes_to_remove)


def _remove_unused_instances_flat(flat_netlist: NetlistDict) -> NetlistDict:
    flat_netlist = {**flat_netlist}

    flat_netlist["instances"] = {**flat_netlist["instances"]}
    flat_netlist["connections"] = {**flat_netlist["connections"]}
    flat_netlist["ports"] = {**flat_netlist["ports"]}

    graph = _get_connectivity_graph(flat_netlist)
    names = set(_get_nodes_to_remove(graph, flat_netlist))

    for name in list(names):
        if name in flat_netlist["instances"]:
            del flat_netlist["instances"][name]

    for conn1, conn2 in list(flat_netlist["connections"].items()):
        for conn in [conn1, conn2]:
            name, _ = conn.split(",")
            if name in names and conn1 in flat_netlist["connections"]:
                del flat_netlist["connections"][conn1]

    for pname, conn in list(flat_netlist["ports"].items()):
        name, _ = conn.split(",")
        if name in names and pname in flat_netlist["ports"]:
            del flat_netlist["ports"][pname]

    return flat_netlist


def _copy_netlist(net):
    net = {
        k: deepcopy(v)
        for k, v in net.items()
        if k in ["instances", "connections", "ports"]
    }
    return net


def _flatten_netlist(recnet, net, sep):
    for name, instance in list(net["instances"].items()):
        component = instance["component"]
        if component not in recnet:
            continue
        del net["instances"][name]
        child_net = _copy_netlist(recnet[component])
        _flatten_netlist(recnet, child_net, sep)
        for iname, iinstance in child_net["instances"].items():
            net["instances"][f"{name}{sep}{iname}"] = iinstance
        ports = {k: f"{name}{sep}{v}" for k, v in child_net["ports"].items()}
        for ip1, ip2 in list(net["connections"].items()):
            n1, p1 = ip1.split(",")
            n2, p2 = ip2.split(",")
            if n1 == name:
                del net["connections"][ip1]
                if p1 not in ports:
                    warnings.warn(
                        f"Port {ip1} not found. Connection {ip1}<->{ip2} ignored."
                    )
                    continue
                net["connections"][ports[p1]] = ip2
            elif n2 == name:
                if p2 not in ports:
                    warnings.warn(
                        f"Port {ip2} not found. Connection {ip1}<->{ip2} ignored."
                    )
                    continue
                net["connections"][ip1] = ports[p2]
        for ip1, ip2 in child_net["connections"].items():
            net["connections"][f"{name}{sep}{ip1}"] = f"{name}{sep}{ip2}"
        for p, ip2 in list(net["ports"].items()):
            try:
                n2, p2 = ip2.split(",")
            except ValueError:
                warnings.warn(f"Unconventional port definition ignored: {p}->{ip2}.")
                continue
            if n2 == name:
                if p2 in ports:
                    net["ports"][p] = ports[p2]
                else:
                    del net["ports"][p]


@overload
def rename_instances(netlist: Netlist, mapping: dict[str, str]) -> Netlist:
    ...


@overload
def rename_instances(
    netlist: RecursiveNetlist, mapping: dict[str, str]
) -> RecursiveNetlist:
    ...


@overload
def rename_instances(netlist: NetlistDict, mapping: dict[str, str]) -> NetlistDict:
    ...


@overload
def rename_instances(
    netlist: RecursiveNetlistDict, mapping: dict[str, str]
) -> RecursiveNetlistDict:
    ...


def rename_instances(
    netlist: Netlist | RecursiveNetlist | NetlistDict | RecursiveNetlistDict,
    mapping: dict[str, str],
) -> Netlist | RecursiveNetlist | NetlistDict | RecursiveNetlistDict:
    given_as_dict = isinstance(netlist, dict)

    if is_recursive(netlist):
        netlist = RecursiveNetlist.model_validate(netlist)
    else:
        netlist = Netlist.model_validate(netlist)

    if isinstance(netlist, RecursiveNetlist):
        net = RecursiveNetlist(
            **{
                k: rename_instances(v, mapping).model_dump()
                for k, v in netlist.root.items()
            }
        )
        return net if not given_as_dict else net.model_dump()

    # it's a sax.Netlist now:
    inverse_mapping = {v: k for k, v in mapping.items()}
    if len(inverse_mapping) != len(mapping):
        raise ValueError("Duplicate names to map onto found.")
    instances = {mapping.get(k, k): v for k, v in netlist.instances.items()}
    connections = {}
    for ip1, ip2 in netlist.connections.items():
        i1, p1 = ip1.split(",")
        i2, p2 = ip2.split(",")
        i1 = mapping.get(i1, i1)
        i2 = mapping.get(i2, i2)
        connections[f"{i1},{p1}"] = f"{i2},{p2}"
    ports = {}
    for q, ip in netlist.ports.items():
        i, p = ip.split(",")
        i = mapping.get(i, i)
        ports[q] = f"{i},{p}"

    placements = {mapping.get(k, k): v for k, v in netlist.placements.items()}
    net = Netlist(
        instances=instances,
        connections=connections,
        ports=ports,
        placements=placements,
        settings=netlist.settings,
    )
    return net if not given_as_dict else net.model_dump()


@overload
def rename_models(netlist: Netlist, mapping: dict[str, str]) -> Netlist:
    ...


@overload
def rename_models(
    netlist: RecursiveNetlist, mapping: dict[str, str]
) -> RecursiveNetlist:
    ...


@overload
def rename_models(netlist: NetlistDict, mapping: dict[str, str]) -> NetlistDict:
    ...


@overload
def rename_models(
    netlist: RecursiveNetlistDict, mapping: dict[str, str]
) -> RecursiveNetlistDict:
    ...


def rename_models(
    netlist: Netlist | RecursiveNetlist | NetlistDict | RecursiveNetlistDict,
    mapping: dict[str, str],
) -> Netlist | RecursiveNetlist | NetlistDict | RecursiveNetlistDict:
    given_as_dict = isinstance(netlist, dict)

    if is_recursive(netlist):
        netlist = RecursiveNetlist.model_validate(netlist)
    else:
        netlist = Netlist.model_validate(netlist)

    if isinstance(netlist, RecursiveNetlist):
        net = RecursiveNetlist(
            **{
                k: rename_models(v, mapping).model_dump()
                for k, v in netlist.root.items()
            }
        )
        return net if not given_as_dict else net.model_dump()

    # it's a sax.Netlist now:
    inverse_mapping = {v: k for k, v in mapping.items()}
    if len(inverse_mapping) != len(mapping):
        raise ValueError("Duplicate names to map onto found.")

    instances = {}
    for k, instance in netlist.instances.items():
        given_as_str = False
        if isinstance(instance, str):
            given_as_str = True
            instance = {
                "component": instance,
                "settings": {},
            }
        elif isinstance(instance, Component):
            instance = instance.model_dump()
        assert isinstance(instance, dict)
        instance["component"] = mapping.get(
            instance["component"], instance["component"]
        )
        if given_as_str:
            instances[k] = instance["component"]
        else:
            instances[k] = instance

    net = Netlist(
        instances=instances,
        connections=netlist.connections,
        ports=netlist.ports,
        placements=netlist.placements,
        settings=netlist.settings,
    )
    return net if not given_as_dict else net.model_dump()



""" SAX Additive Backend """

from __future__ import annotations

from typing import Any, Dict, Tuple

import jax.numpy as jnp
import networkx as nx

from ..netlist import Component
from ..saxtypes import Model, SDict, sdict


def analyze_instances_additive(
    instances: Dict[str, Component],
    models: Dict[str, Model],
) -> Dict[str, SDict]:
    instances, instances_old = {}, instances
    for k, v in instances_old.items():
        if not isinstance(v, Component):
            v = Component(**v)
        instances[k] = v
    model_names = set()
    for i in instances.values():
        model_names.add(i.component)
    dummy_models = {k: sdict(models[k]()) for k in model_names}
    dummy_instances = {}
    for k, i in instances.items():
        dummy_instances[k] = dummy_models[i.component]
    return dummy_instances


def analyze_circuit_additive(
    analyzed_instances: Dict[str, SDict],
    connections: Dict[str, str],
    ports: Dict[str, str],
) -> Any:
    return connections, ports


def evaluate_circuit_additive(
    analyzed: Any,
    instances: Dict[str, SDict],
) -> SDict:
    """evaluate a circuit for the given sdicts."""
    connections, ports = analyzed
    edges = _graph_edges(instances, connections, ports)

    graph = nx.Graph()
    graph.add_edges_from(edges)
    _prune_internal_output_nodes(graph)

    sdict = {}
    for source in ports:
        for target in ports:
            paths = _get_possible_paths(graph, source=("", source), target=("", target))
            if not paths:
                continue
            sdict[source, target] = _path_lengths(graph, paths)

    return sdict


def _split_port(port: str) -> Tuple[str, str]:
    try:
        instance, port = port.split(",")
    except ValueError:
        (port,) = port.split(",")
        instance = ""
    return instance, port


def _graph_edges(
    instances: Dict[str, SDict],
    connections: Dict[str, str],
    ports: Dict[str, str],
):
    zero = jnp.array([0.0], dtype=float)
    edges = {}
    edges.update({_split_port(k): _split_port(v) for k, v in connections.items()})
    edges.update({_split_port(v): _split_port(k) for k, v in connections.items()})
    edges.update({_split_port(k): _split_port(v) for k, v in ports.items()})
    edges.update({_split_port(v): _split_port(k) for k, v in ports.items()})
    edges = [(n1, n2, {"type": "C", "length": zero}) for n1, n2 in edges.items()]

    _instances = {
        **{i1: None for (i1, _), (_, _), _ in edges},
        **{i2: None for (_, _), (i2, _), _ in edges},
    }

    if "" in _instances:
        del _instances[""]  # external ports don't belong to an instance

    for instance in _instances:
        s = instances[instance]
        edges += [
            (
                (instance, p1),
                (instance, p2),
                {"type": "S", "length": jnp.asarray(length, dtype=float).ravel()},
            )
            for (p1, p2), length in sdict(s).items()
        ]

    return edges


def _prune_internal_output_nodes(graph):
    broken = True
    while broken:
        broken = False
        for (i, p), dic in list(graph.adjacency()):
            if (
                i != ""
                and len(dic) == 2
                and all(prop.get("type", "C") == "C" for prop in dic.values())
            ):
                graph.remove_node((i, p))
                graph.add_edge(*dic.keys(), type="C", length=0.0)
                broken = True
                break
    return graph


def _get_possible_paths(graph, source, target):
    paths = []
    default_props = {"type": "C", "length": 0.0}
    for path in nx.all_simple_edge_paths(graph, source, target):
        prevtype = "C"
        for n1, n2 in path:
            curtype = graph.get_edge_data(n1, n2, default_props)["type"]
            if curtype == prevtype == "S":
                break
            else:
                prevtype = curtype
        else:
            paths.append(path)
    return paths


def _path_lengths(graph, paths):
    lengths = []
    for path in paths:
        length = zero = jnp.array([0.0], dtype=float)
        default_edge_data = {"type": "C", "length": zero}
        for edge in path:
            edge_data = graph.get_edge_data(*edge, default_edge_data)
            length = (length[None, :] + edge_data.get("length", zero)[:, None]).ravel()
        lengths.append(length)
    return lengths



""" SAX Circuit Definition """

from __future__ import annotations

import os
import shutil
import sys
from functools import partial
from typing import Callable, NamedTuple

import black
import networkx as nx
import numpy as np

from .backends import circuit_backends, backend_map
from .netlist import AnyNetlist, Netlist, NetlistDict, RecursiveNetlist, is_recursive
from .netlist import netlist as parse_netlist
from .saxtypes import Model, Settings, SType, scoo, sdense, sdict
from .utils import (
    _replace_kwargs,
    get_ports,
    get_settings,
    merge_dicts,
    update_settings,
)


class CircuitInfo(NamedTuple):
    """Information about the circuit function you created."""

    dag: nx.DiGraph[str]
    models: dict[str, Model]
    backend: str


def circuit(
    netlist: AnyNetlist,
    models: dict[str, Model] | None = None,
    backend: str = "default",
    return_type: str = "sdict",
    ignore_missing_ports: bool = False,
) -> tuple[Model, CircuitInfo]:
    """Create a circuit function for a given netlist.

    Args:
        netlist: The netlist to create a circuit for.
        models: A dictionary of models to use in the circuit.
        backend: The backend to use for the circuit.
        return_type: The type of the circuit function to return.
        ignore_missing_ports: Ignore missing ports in the netlist.

    """
    backend = _validate_circuit_backend(backend)

    instance_models = _extract_instance_models(netlist)
    recnet: RecursiveNetlist = parse_netlist(
        netlist, with_unconnected_instances=False, with_placements=False
    )
    _validate_netlist_ports(recnet)
    dependency_dag: nx.DiGraph[str] = _create_dag(recnet, models, validate=True)
    models = _validate_models(models, dependency_dag, extra_models=instance_models)

    circuit = None
    new_models = {}
    current_models = {}
    model_names = list(nx.topological_sort(dependency_dag))[::-1]
    for model_name in model_names:
        if model_name in models:
            new_models[model_name] = models[model_name]
            continue

        flatnet = recnet.root[model_name]
        current_models |= new_models
        new_models = {}

        current_models[model_name] = circuit = _flat_circuit(
            flatnet.instances,
            flatnet.connections,
            flatnet.ports,
            current_models,
            backend,
            ignore_missing_ports=ignore_missing_ports,
        )

    assert circuit is not None
    circuit = _enforce_return_type(circuit, return_type)
    return circuit, CircuitInfo(
        dag=dependency_dag,
        models=current_models,
        backend=backend_map.get(backend, backend),
    )


def _create_dag(
    netlist: RecursiveNetlist,
    models: dict[str, Model] | None = None,
    validate: bool = False,
) -> nx.DiGraph[str]:
    if models is None:
        models = {}
    assert isinstance(models, dict)

    all_models = {}
    g = nx.DiGraph()

    for model_name, subnetlist in netlist.model_dump().items():
        if model_name not in all_models:
            all_models[model_name] = models.get(model_name, subnetlist)
            g.add_node(model_name)
        if model_name in models:
            continue
        for instance in subnetlist["instances"].values():
            component = instance["component"]
            if component not in all_models:
                all_models[component] = models.get(component, None)
                g.add_node(component)
            g.add_edge(model_name, component)

    # we only need the nodes that depend on the parent...
    parent_node = next(iter(netlist.root.keys()))
    nodes = [parent_node, *nx.descendants(g, parent_node)]
    g = nx.induced_subgraph(g, nodes)
    assert isinstance(g, nx.DiGraph)
    if validate:
        g = _validate_dag(g)
    return g


def draw_dag(dag, with_labels=True, **kwargs):
    _patch_path()
    if shutil.which("dot"):
        return nx.draw(
            dag,
            nx.nx_pydot.pydot_layout(dag, prog="dot"),
            with_labels=with_labels,
            **kwargs,
        )
    else:
        return nx.draw(dag, _my_dag_pos(dag), with_labels=with_labels, **kwargs)


def get_required_circuit_models(
    netlist: AnyNetlist,
    models: dict[str, Model] | None = None,
) -> list[str]:
    """Figure out which models are needed for a given netlist.

    Args:
        netlist: The netlist to create a circuit for.
        models: A dictionary of models to use in the circuit.

    """
    instance_models = _extract_instance_models(netlist)
    recnet: RecursiveNetlist = parse_netlist(
        netlist, with_unconnected_instances=False, with_placements=False
    )
    dependency_dag: nx.DiGraph[str] = _create_dag(recnet, models, validate=True)
    _, required, _ = _find_missing_models(
        models, dependency_dag, extra_models=instance_models
    )
    return required


def _flat_circuit(
    instances, connections, ports, models, backend, ignore_missing_ports=False
):
    analyze_insts_fn, analyze_fn, evaluate_fn = circuit_backends[backend]
    dummy_instances = analyze_insts_fn(instances, models)
    inst_port_mode = {
        k: _port_modes_dict(get_ports(s)) for k, s in dummy_instances.items()
    }
    connections = _get_multimode_connections(
        connections, inst_port_mode, ignore_missing_ports=ignore_missing_ports
    )
    ports = _get_multimode_ports(
        ports, inst_port_mode, ignore_missing_ports=ignore_missing_ports
    )

    inst2model = {}
    for k, inst in instances.items():
        inst2model[k] = models[inst.component]

    model_settings = {name: get_settings(model) for name, model in inst2model.items()}
    netlist_settings = {
        name: {
            k: v for k, v in (inst.settings or {}).items() if k in model_settings[name]
        }
        for name, inst in instances.items()
    }
    default_settings = merge_dicts(model_settings, netlist_settings)
    analyzed = analyze_fn(dummy_instances, connections, ports)

    def _circuit(**settings: Settings) -> SType:
        full_settings = merge_dicts(default_settings, settings)
        full_settings = _forward_global_settings(inst2model, full_settings)
        full_settings = merge_dicts(full_settings, settings)

        instances: dict[str, SType] = {}
        for inst_name, model in inst2model.items():
            instances[inst_name] = model(**full_settings.get(inst_name, {}))

        S = evaluate_fn(analyzed, instances)
        return S

    _replace_kwargs(_circuit, **default_settings)

    return _circuit


def _patch_path():
    os_paths = {p: None for p in os.environ.get("PATH", "").split(os.pathsep)}
    sys_paths = {p: None for p in sys.path}
    other_paths = {os.path.dirname(sys.executable): None}
    os.environ["PATH"] = os.pathsep.join(os_paths | sys_paths | other_paths)


def _my_dag_pos(dag):
    # inferior to pydot
    in_degree = {}
    for k, v in dag.in_degree():
        if v not in in_degree:
            in_degree[v] = []
        in_degree[v].append(k)

    widths = {k: len(vs) for k, vs in in_degree.items()}
    width = max(widths.values())

    horizontal_pos = {
        k: np.linspace(0, 1, w + 2)[1:-1] * width for k, w in widths.items()
    }

    pos = {}
    for k, vs in in_degree.items():
        for x, v in zip(horizontal_pos[k], vs):
            pos[v] = (x, -k)
    return pos


def _find_root(g):
    return [n for n, d in g.in_degree() if d == 0]


def _find_leaves(g):
    return [n for n, d in g.out_degree() if d == 0]


def _find_missing_models(
    models: dict | None, dag: nx.DiGraph, extra_models: dict | None = None
) -> tuple[dict[str, Callable], list[str], list[str]]:
    if extra_models is None:
        extra_models = {}
    if models is None:
        models = {}
    models = {**models, **extra_models}
    required_models = _find_leaves(dag)
    missing_models = [m for m in required_models if m not in models]
    return models, required_models, missing_models


def _validate_models(
    models: dict | None, dag: nx.DiGraph, extra_models: dict | None = None
) -> dict[str, Model]:
    models, required_models, missing_models = _find_missing_models(
        models, dag, extra_models
    )
    if missing_models:
        model_diff = {
            "Missing Models": missing_models,
            "Given Models": list(models),
            "Required Models": required_models,
        }
        raise ValueError(
            "Missing models. The following models are still missing to build "
            f"the circuit:\n{black.format_str(repr(model_diff), mode=black.Mode())}"
        )
    return models


def _forward_global_settings(instances, settings):
    global_settings = {
        k: settings.pop(k) for k in list(settings.keys()) if k not in instances
    }
    if global_settings:
        settings = update_settings(settings, **global_settings)
    return settings


def _port_modes_dict(port_modes):
    result = {}
    for port_mode in port_modes:
        port, mode = port_mode.split("@") if "@" in port_mode else (port_mode, None)
        if port not in result:
            result[port] = set()
        if mode is not None:
            result[port].add(mode)
    return result


def _get_multimode_connections(connections, inst_port_mode, ignore_missing_ports=False):
    mm_connections = {}
    for inst_port1, inst_port2 in connections.items():
        inst1, port1 = inst_port1.split(",")
        inst2, port2 = inst_port2.split(",")
        try:
            modes1 = inst_port_mode[inst1][port1]
        except KeyError:
            if ignore_missing_ports:
                continue
            raise RuntimeError(
                f"Instance {inst1} does not contain port {port1}. Available ports: {list(inst_port_mode[inst1])}."
            )
        try:
            modes2 = inst_port_mode[inst2][port2]
        except KeyError:
            if ignore_missing_ports:
                continue
            raise RuntimeError(
                f"Instance {inst2} does not contain port {port2}. Available ports: {list(inst_port_mode[inst2])}."
            )
        if not modes1 and not modes2:
            mm_connections[f"{inst1},{port1}"] = f"{inst2},{port2}"
        elif (not modes1) or (not modes2):
            raise ValueError(
                "trying to connect a multimode model to single mode model.\n"
                "Please update your models dictionary.\n"
                f"Problematic connection: '{inst_port1}':'{inst_port2}'"
            )
        else:
            common_modes = modes1.intersection(modes2)
            for mode in sorted(common_modes):
                mm_connections[f"{inst1},{port1}@{mode}"] = f"{inst2},{port2}@{mode}"
    return mm_connections


def _get_multimode_ports(ports, inst_port_mode, ignore_missing_ports=False):
    mm_ports = {}
    for port, inst_port2 in ports.items():
        inst2, port2 = inst_port2.split(",")
        try:
            modes2 = inst_port_mode[inst2][port2]
        except KeyError:
            if ignore_missing_ports:
                continue
            raise RuntimeError(
                f"Instance {inst2} does not contain port {port2}. Available ports: {list(inst_port_mode[inst2])}"
            )
        if not modes2:
            mm_ports[port] = f"{inst2},{port2}"
        else:
            for mode in sorted(modes2):
                mm_ports[f"{port}@{mode}"] = f"{inst2},{port2}@{mode}"
    return mm_ports


def _enforce_return_type(model, return_type):
    stype_func = {
        "default": lambda x: x,
        "stype": lambda x: x,
        "sdict": sdict,
        "scoo": scoo,
        "sdense": sdense,
    }[return_type]
    return stype_func(model)


def _extract_instance_models(netlist: AnyNetlist) -> dict[str, Model]:
    if isinstance(netlist, Netlist):
        return {}
    elif isinstance(netlist, RecursiveNetlist):
        return {}
    elif isinstance(netlist, dict):
        if is_recursive(netlist):
            models = {}
            for net in netlist.values():
                models.update(_extract_instance_models(net))  # type: ignore
            return models
        else:
            callable_instances = [
                f for f in netlist["instances"].values() if callable(f)
            ]
            models = {}
            for f in callable_instances:
                while isinstance(f, partial):
                    f = f.func
                models[f.__name__] = f
            return models
    else:
        return {}


def _validate_circuit_backend(backend):
    backend = backend.lower()
    # assert valid circuit_backend
    if backend not in circuit_backends:
        raise KeyError(
            f"circuit backend {backend} not found. Allowed circuit backends: "
            f"{', '.join(circuit_backends.keys())}."
        )
    return backend


def _validate_dag(dag):
    nodes = _find_root(dag)
    if len(nodes) > 1:
        raise ValueError(f"Multiple top_levels found in netlist: {nodes}")
    if len(nodes) < 1:
        raise ValueError("Netlist does not contain any nodes.")
    if not dag.is_directed():
        raise ValueError("Netlist dependency cycles detected!")
    return dag


def _validate_netlist_ports(netlist: RecursiveNetlist):
    if len(netlist.root) < 1:
        raise ValueError("Cannot create circuit: empty netlist")
    net: Netlist = netlist.root[list(netlist.root)[0]]
    ports_str = ", ".join(list(net.ports))
    if not ports_str:
        ports_str = "no ports given"
    if len(net.ports) < 2:
        raise ValueError(
            "Cannot create circuit: "
            f"at least 2 ports need to be defined. Got {ports_str}."
        )



""" SAX Default Models """

from __future__ import annotations

from functools import lru_cache as cache
from typing import Optional, Tuple

import jax
import jax.numpy as jnp

from .saxtypes import FloatArrayND, Model, SCoo, SDict
from .utils import get_inputs_outputs, reciprocal


def straight(
    *,
    wl: FloatArrayND | float = 1.55,
    wl0: float = 1.55,
    neff: float = 2.34,
    ng: float = 3.4,
    length: float = 10.0,
    loss: float = 0.0,
) -> SDict:
    """A simple straight waveguide model.

    Args:
        wl: wavelength in microns.
        wl0: reference wavelength in microns.
        neff: effective index.
        ng: group index.
        length: length of the waveguide in microns.
        loss: loss in dB/cm.

    """
    dwl = wl - wl0
    dneff_dwl = (ng - neff) / wl0
    _neff = neff - dwl * dneff_dwl
    phase = 2 * jnp.pi * _neff * length / wl
    amplitude = jnp.asarray(10 ** (-loss * length / 20), dtype=complex)
    transmission = amplitude * jnp.exp(1j * phase)
    return reciprocal(
        {
            ("in0", "out0"): transmission,
        }
    )


def coupler(*, coupling: float = 0.5) -> SDict:
    """a simple coupler model"""
    kappa = coupling**0.5
    tau = (1 - coupling) ** 0.5
    return reciprocal(
        {
            ("in0", "out0"): tau,
            ("in0", "out1"): 1j * kappa,
            ("in1", "out0"): 1j * kappa,
            ("in1", "out1"): tau,
        }
    )


def _validate_ports(
    ports, num_inputs, num_outputs, diagonal
) -> Tuple[Tuple[str, ...], Tuple[str, ...], int, int]:
    """Validate the ports and return the input and output ports."""

    if ports is None:
        if num_inputs is None or num_outputs is None:
            raise ValueError(
                "if not ports given, you must specify how many input ports "
                "and how many output ports a model has."
            )
        input_ports = [f"in{i}" for i in range(num_inputs)]
        output_ports = [f"out{i}" for i in range(num_outputs)]
    else:
        if num_inputs is not None:
            if num_outputs is None:
                raise ValueError(
                    "if num_inputs is given, num_outputs should be given as well."
                )
        if num_outputs is not None:
            if num_inputs is None:
                raise ValueError(
                    "if num_outputs is given, num_inputs should be given as well."
                )
        if num_inputs is not None and num_outputs is not None:
            if num_inputs + num_outputs != len(ports):
                raise ValueError("num_inputs + num_outputs != len(ports)")
            input_ports = ports[:num_inputs]
            output_ports = ports[num_inputs:]
        else:
            input_ports, output_ports = get_inputs_outputs(ports)
            num_inputs = len(input_ports)
            num_outputs = len(output_ports)

    if diagonal:
        if num_inputs != num_outputs:
            raise ValueError(
                "Can only have a diagonal passthru if number "
                "of input ports equals the number of output ports!"
            )

    return tuple(input_ports), tuple(output_ports), num_inputs, num_outputs


@cache
def unitary(
    num_inputs: Optional[int] = None,
    num_outputs: Optional[int] = None,
    ports: Optional[Tuple[str, ...]] = None,
    *,
    jit=True,
    reciprocal=True,
    diagonal=False,
) -> Model:
    """A unitary model.

    Args:
        num_inputs: number of input ports.
        num_outputs: number of output ports.
        ports: tuple of input and output ports.
        jit: whether to jit the model.
        reciprocal: whether the model is reciprocal.
        diagonal: whether the model is diagonal.

    """
    input_ports, output_ports, num_inputs, num_outputs = _validate_ports(
        ports, num_inputs, num_outputs, diagonal
    )
    assert num_inputs is not None and num_outputs is not None

    # let's create the squared S-matrix:
    N = max(num_inputs, num_outputs)
    S = jnp.zeros((2 * N, 2 * N), dtype=float)

    if not diagonal:
        S = S.at[:N, N:].set(1)
    else:
        r = jnp.arange(
            N, dtype=int
        )  # reciprocal only works if num_inputs == num_outputs!
        S = S.at[r, N + r].set(1)

    if reciprocal:
        if not diagonal:
            S = S.at[N:, :N].set(1)
        else:
            r = jnp.arange(
                N, dtype=int
            )  # reciprocal only works if num_inputs == num_outputs!
            S = S.at[N + r, r].set(1)

    # Now we need to normalize the squared S-matrix
    U, s, V = jnp.linalg.svd(S, full_matrices=False)
    S = jnp.sqrt(U @ jnp.diag(jnp.where(s > 1e-12, 1, 0)) @ V)

    # Now create subset of this matrix we're interested in:
    r = jnp.concatenate(
        [jnp.arange(num_inputs, dtype=int), N + jnp.arange(num_outputs, dtype=int)], 0
    )
    S = S[r, :][:, r]

    # let's convert it in SCOO format:
    Si, Sj = jnp.where(S > 1e-6)
    Sx = S[Si, Sj]

    # the last missing piece is a port map:
    pm = {
        **{p: i for i, p in enumerate(input_ports)},
        **{p: i + num_inputs for i, p in enumerate(output_ports)},
    }

    def func(wl: float = 1.5) -> SCoo:
        wl_ = jnp.asarray(wl)
        Sx_ = jnp.broadcast_to(Sx, (*wl_.shape, *Sx.shape))
        return Si, Sj, Sx_, pm

    func.__name__ = f"unitary_{num_inputs}_{num_outputs}"
    func.__qualname__ = f"unitary_{num_inputs}_{num_outputs}"
    return jax.jit(func) if jit else func


@cache
def copier(
    num_inputs: Optional[int] = None,
    num_outputs: Optional[int] = None,
    ports: Optional[Tuple[str, ...]] = None,
    *,
    jit=True,
    reciprocal=True,
    diagonal=False,
) -> Model:
    """A copier model.

    Args:
        num_inputs: number of input ports.
        num_outputs: number of output ports.
        ports: tuple of input and output ports.
        jit: whether to jit the model.
        reciprocal: whether the model is reciprocal.
        diagonal: whether the model is diagonal.
    """
    input_ports, output_ports, num_inputs, num_outputs = _validate_ports(
        ports, num_inputs, num_outputs, diagonal
    )
    assert num_inputs is not None and num_outputs is not None

    # let's create the squared S-matrix:
    S = jnp.zeros((num_inputs + num_outputs, num_inputs + num_outputs), dtype=float)

    if not diagonal:
        S = S.at[:num_inputs, num_inputs:].set(1)
    else:
        r = jnp.arange(
            num_inputs, dtype=int
        )  # == range(num_outputs) # reciprocal only works if num_inputs == num_outputs!
        S = S.at[r, num_inputs + r].set(1)

    if reciprocal:
        if not diagonal:
            S = S.at[num_inputs:, :num_inputs].set(1)
        else:
            # reciprocal only works if num_inputs == num_outputs!
            r = jnp.arange(num_inputs, dtype=int)  # == range(num_outputs)
            S = S.at[num_inputs + r, r].set(1)

    # let's convert it in SCOO format:
    Si, Sj = jnp.where(S > 1e-6)
    Sx = S[Si, Sj]

    # the last missing piece is a port map:
    pm = {
        **{p: i for i, p in enumerate(input_ports)},
        **{p: i + num_inputs for i, p in enumerate(output_ports)},
    }

    def func(wl: float = 1.5) -> SCoo:
        wl_ = jnp.asarray(wl)
        Sx_ = jnp.broadcast_to(Sx, (*wl_.shape, *Sx.shape))
        return Si, Sj, Sx_, pm

    func.__name__ = f"unitary_{num_inputs}_{num_outputs}"
    func.__qualname__ = f"unitary_{num_inputs}_{num_outputs}"
    return jax.jit(func) if jit else func


@cache
def passthru(
    num_links: Optional[int] = None,
    ports: Optional[Tuple[str, ...]] = None,
    *,
    jit=True,
    reciprocal=True,
) -> Model:
    """A passthru model.

    Args:
        num_links: number of links.
        ports: tuple of input and output ports.
        jit: whether to jit the model.
        reciprocal: whether the model is reciprocal.
    """
    passthru = unitary(
        num_links, num_links, ports, jit=jit, reciprocal=reciprocal, diagonal=True
    )
    passthru.__name__ = f"passthru_{num_links}_{num_links}"
    passthru.__qualname__ = f"passthru_{num_links}_{num_links}"
    return jax.jit(passthru) if jit else passthru


models = {
    "copier": copier,
    "coupler": coupler,
    "passthru": passthru,
    "straight": straight,
    "unitary": unitary,
}


def get_models(copy: bool = True):
    return {**models} if copy else models



import sax

SAMPLE_NETLIST = {
    "instances": {
        "lft": "coupler",
        "top": "waveguide",
        "btm": "waveguide",
        "rgt": "coupler",
    },
    "connections": {
        "lft,out0": "btm,in0",
        "btm,out0": "rgt,in0",
        "lft,out1": "top,in0",
        "top,out0": "rgt,in1",
    },
    "ports": {
        "in0": "lft,in0",
        "in1": "lft,in1",
        "out0": "rgt,out0",
        "out1": "rgt,out1",
    },
}


def test_circuit():
    mzi, info = sax.circuit(
        netlist=SAMPLE_NETLIST,
        models={
            "coupler": sax.models.coupler,
            "waveguide": sax.models.straight,
        },
    )


if __name__ == "__main__":
    pass



"""SAX KLU Backend"""

from __future__ import annotations

import time
from typing import Any, Dict

import jax
import jax.numpy as jnp
from klujax import solve as _solve_klu
from natsort import natsorted

from ..netlist import Component
from ..saxtypes import Model, SCoo, SDense, SType, scoo

try:
    from klujax import dot as _dot
except ImportError:
    from klujax import coo_mul_vec as _dot

solve_klu = jax.vmap(_solve_klu, (None, None, 0, None), 0)
mul_coo = jax.vmap(_dot, (None, None, 0, 0), 0)


def analyze_instances_klu(
    instances: Dict[str, Component],
    models: Dict[str, Model],
) -> Dict[str, SCoo]:
    instances, instances_old = {}, instances
    for k, v in instances_old.items():
        if not isinstance(v, Component):
            v = Component(**v)
        instances[k] = v
    model_names = set()
    for i in instances.values():
        model_names.add(i.component)
    dummy_models = {k: scoo(models[k]()) for k in model_names}
    dummy_instances = {}
    for k, i in instances.items():
        dummy_instances[k] = dummy_models[i.component]
    return dummy_instances


def analyze_circuit_klu(
    analyzed_instances: Dict[str, SCoo],
    connections: Dict[str, str],
    ports: Dict[str, str],
) -> Any:
    connections = {**connections, **{v: k for k, v in connections.items()}}
    inverse_ports = {v: k for k, v in ports.items()}
    port_map = {k: i for i, k in enumerate(ports)}

    idx, Si, Sj, instance_ports = 0, [], [], {}
    for name, instance in analyzed_instances.items():
        si, sj, _, ports_map = scoo(instance)
        Si.append(si + idx)
        Sj.append(sj + idx)
        instance_ports.update({f"{name},{p}": i + idx for p, i in ports_map.items()})
        idx += len(ports_map)

    n_col = idx
    n_rhs = len(port_map)

    Si = jnp.concatenate(Si, -1)
    Sj = jnp.concatenate(Sj, -1)

    Cmap = {
        int(instance_ports[k]): int(instance_ports[v]) for k, v in connections.items()
    }
    Ci = jnp.array(list(Cmap.keys()), dtype=jnp.int32)
    Cj = jnp.array(list(Cmap.values()), dtype=jnp.int32)

    Cextmap = {
        int(instance_ports[k]): int(port_map[v]) for k, v in inverse_ports.items()
    }
    Cexti = jnp.stack(list(Cextmap.keys()), 0)
    Cextj = jnp.stack(list(Cextmap.values()), 0)
    Cext = jnp.zeros((n_col, n_rhs), dtype=complex).at[Cexti, Cextj].set(1.0)

    mask = Cj[None, :] == Si[:, None]
    CSi = jnp.broadcast_to(Ci[None, :], mask.shape)[mask]

    mask = (Cj[:, None] == Si[None, :]).any(0)
    CSj = Sj[mask]

    Ii = Ij = jnp.arange(n_col)
    I_CSi = jnp.concatenate([CSi, Ii], -1)
    I_CSj = jnp.concatenate([CSj, Ij], -1)
    return (
        n_col,
        mask,
        Si,
        Sj,
        Cext,
        Cexti,
        Cextj,
        I_CSi,
        I_CSj,
        tuple((k, v[1]) for k, v in analyzed_instances.items()),
        tuple(port_map),
    )


def evaluate_circuit_klu(analyzed: Any, instances: Dict[str, SType]) -> SDense:
    (
        n_col,
        mask,
        Si,
        Sj,
        Cext,
        Cexti,
        Cextj,
        I_CSi,
        I_CSj,
        dummy_pms,
        port_map,
    ) = analyzed

    idx = 0
    Sx = []
    batch_shape = ()
    for name, pm_ in dummy_pms:
        _, _, sx, ports_map = scoo(instances[name])
        Sx.append(sx)
        if len(sx.shape[:-1]) > len(batch_shape):
            batch_shape = sx.shape[:-1]
        idx += len(ports_map)

    Sx = jnp.concatenate(
        [jnp.broadcast_to(sx, (*batch_shape, sx.shape[-1])) for sx in Sx], -1
    )
    CSx = Sx[..., mask]
    Ix = jnp.ones((*batch_shape, n_col))
    I_CSx = jnp.concatenate([-CSx, Ix], -1)

    Sx = Sx.reshape(-1, Sx.shape[-1])  # n_lhs x N
    I_CSx = I_CSx.reshape(-1, I_CSx.shape[-1])  # n_lhs x M
    inv_I_CS_Cext = solve_klu(I_CSi, I_CSj, I_CSx, Cext)
    S_inv_I_CS_Cext = mul_coo(Si, Sj, Sx, inv_I_CS_Cext)

    CextT_S_inv_I_CS_Cext = S_inv_I_CS_Cext[..., Cexti, :][..., :, Cextj]

    _, n, _ = CextT_S_inv_I_CS_Cext.shape
    S = CextT_S_inv_I_CS_Cext.reshape(*batch_shape, n, n)

    return S, {p: i for i, p in enumerate(port_map)}


def _get_instance_ports(connections: Dict[str, str], ports: Dict[str, str]):
    instance_ports = {}
    for connection in connections.items():
        for ip in connection:
            i, p = ip.split(",")
            if i not in instance_ports:
                instance_ports[i] = set()
            instance_ports[i].add(p)
    for ip in ports.values():
        i, p = ip.split(",")
        if i not in instance_ports:
            instance_ports[i] = set()
        instance_ports[i].add(p)
    return {k: natsorted(v) for k, v in instance_ports.items()}


def _get_dummy_instances(connections, ports):
    """no longer used. deprecated by analyze_instances_klu."""
    instance_ports = _get_instance_ports(connections, ports)
    dummy_instances = {}
    for name, ports in instance_ports.items():
        num_ports = len(ports)
        pm = {p: i for i, p in enumerate(ports)}
        ij = jnp.mgrid[:num_ports, :num_ports]
        i = ij[0].ravel()
        j = ij[1].ravel()
        dummy_instances[name] = (i, j, None, pm)
    return dummy_instances



""" SAX Filipsson Gunnar Backend """

from __future__ import annotations

from typing import Any, Dict

import jax

from ..netlist import Component
from ..saxtypes import Model, SDict, SType, sdict


def analyze_instances_fg(
    instances: Dict[str, Component],
    models: Dict[str, Model],
) -> Dict[str, SDict]:
    instances, instances_old = {}, instances
    for k, v in instances_old.items():
        if not isinstance(v, Component):
            v = Component(**v)
        instances[k] = v
    model_names = set()
    for i in instances.values():
        model_names.add(i.component)
    dummy_models = {k: sdict(models[k]()) for k in model_names}
    dummy_instances = {}
    for k, i in instances.items():
        dummy_instances[k] = dummy_models[i.component]
    return dummy_instances


def analyze_circuit_fg(
    analyzed_instances: Dict[str, SDict],
    connections: Dict[str, str],
    ports: Dict[str, str],
) -> Any:
    # skip analysis for now
    return connections, ports


def evaluate_circuit_fg(
    analyzed: Any,
    instances: Dict[str, SType],
) -> SDict:
    """evaluate a circuit for the given sdicts."""
    connections, ports = analyzed

    # it's actually easier working w reverse:
    reversed_ports = {v: k for k, v in ports.items()}

    block_diag = {}
    for name, S in instances.items():
        block_diag.update(
            {(f"{name},{p1}", f"{name},{p2}"): v for (p1, p2), v in sdict(S).items()}
        )

    sorted_connections = sorted(connections.items(), key=_connections_sort_key)
    all_connected_instances = {k: {k} for k in instances}

    for k, l in sorted_connections:  # noqa: E741
        name1, _ = k.split(",")
        name2, _ = l.split(",")

        connected_instances = (
            all_connected_instances[name1] | all_connected_instances[name2]
        )
        for name in connected_instances:
            all_connected_instances[name] = connected_instances

        current_ports = tuple(
            p
            for instance in connected_instances
            for p in set([p for p, _ in block_diag] + [p for _, p in block_diag])
            if p.startswith(f"{instance},")
        )

        block_diag.update(_interconnect_ports(block_diag, current_ports, k, l))

        for i, j in list(block_diag.keys()):
            is_connected = i == k or i == l or j == k or j == l
            is_in_output_ports = i in reversed_ports and j in reversed_ports
            if is_connected and not is_in_output_ports:
                del block_diag[
                    i, j
                ]  # we're no longer interested in these port combinations

    circuit_sdict: SDict = {
        (reversed_ports[i], reversed_ports[j]): v
        for (i, j), v in block_diag.items()
        if i in reversed_ports and j in reversed_ports
    }
    return circuit_sdict


def _connections_sort_key(connection):
    """sort key for sorting a connection dictionary"""
    part1, part2 = connection
    name1, _ = part1.split(",")
    name2, _ = part2.split(",")
    return (min(name1, name2), max(name1, name2))


def _interconnect_ports(block_diag, current_ports, k, l):  # noqa: E741
    """interconnect two ports in a given model

    > Note: the interconnect algorithm is based on equation 6 of 'Filipsson, Gunnar.
      "A new general computer algorithm for S-matrix calculation of interconnected
      multiports." 11th European Microwave Conference. IEEE, 1981.'
    """
    current_block_diag = {}
    for i in current_ports:
        for j in current_ports:
            vij = _calculate_interconnected_value(
                vij=block_diag.get((i, j), 0.0),
                vik=block_diag.get((i, k), 0.0),
                vil=block_diag.get((i, l), 0.0),
                vkj=block_diag.get((k, j), 0.0),
                vkk=block_diag.get((k, k), 0.0),
                vkl=block_diag.get((k, l), 0.0),
                vlj=block_diag.get((l, j), 0.0),
                vlk=block_diag.get((l, k), 0.0),
                vll=block_diag.get((l, l), 0.0),
            )
            current_block_diag[i, j] = vij
    return current_block_diag


@jax.jit
def _calculate_interconnected_value(vij, vik, vil, vkj, vkk, vkl, vlj, vlk, vll):
    """Calculate an interconnected S-parameter value

    Note:
        The interconnect algorithm is based on equation 6 in the paper below::

          Filipsson, Gunnar. "A new general computer algorithm for S-matrix calculation
          of interconnected multiports." 11th European Microwave Conference. IEEE, 1981.
    """
    result = vij + (
        vkj * vil * (1 - vlk)
        + vlj * vik * (1 - vkl)
        + vkj * vll * vik
        + vlj * vkk * vil
    ) / ((1 - vkl) * (1 - vlk) - vkk * vll)
    return result



""" SAX Loss Functions """

from __future__ import annotations

from typing import Dict, cast

import jax.numpy as jnp

from .saxtypes import ComplexArrayND


def mse(x: ComplexArrayND, y: ComplexArrayND) -> float:
    """mean squared error"""
    return cast(float, (abs(x - y) ** 2).mean())


def huber_loss(x: ComplexArrayND, y: ComplexArrayND, delta: float = 0.5) -> float:
    """huber loss"""
    loss = ((delta**2) * ((1.0 + (abs(x - y) / delta) ** 2) ** 0.5 - 1.0)).mean()
    return cast(float, loss)


def l2_reg(weights: Dict[str, ComplexArrayND]) -> float:
    """L2 regularization loss"""
    numel = 0
    loss = 0.0
    for w in (v for k, v in weights.items() if k[0] in ("w", "b")):
        numel = numel + w.size
        loss = loss + (jnp.abs(w) ** 2).sum()
    return cast(float, loss / numel)



"""SAX: S + Autograd + XLA"""

from __future__ import annotations

__author__ = "Floris Laporte"
__version__ = "0.14.4"


from functools import partial as partial
from math import pi as pi

from scipy.constants import c as c

from . import backends as backends
from . import models as models
from . import saxtypes as saxtypes
from . import utils as utils
from .circuit import circuit as circuit
from .circuit import get_required_circuit_models as get_required_circuit_models
from .interpolation import interpolate_xarray as interpolate_xarray
from .loss import huber_loss as huber_loss
from .loss import l2_reg as l2_reg
from .loss import mse as mse
from .models import get_models as get_models
from .models import passthru as passthru
from .multimode import multimode as multimode
from .multimode import singlemode as singlemode
from .netlist import Netlist as Netlist
from .netlist import RecursiveNetlist as RecursiveNetlist
from .netlist import flatten_netlist as flatten_netlist
from .netlist import get_component_instances as get_component_instances
from .netlist import get_netlist_instances_by_prefix as get_netlist_instances_by_prefix
from .netlist import load_netlist as load_netlist
from .netlist import load_recursive_netlist as load_recursive_netlist
from .netlist import netlist as netlist
from .netlist import rename_instances as rename_instances
from .netlist import rename_models as rename_models
from .saxtypes import Array as Array
from .saxtypes import ArrayLike as ArrayLike
from .saxtypes import Complex as Complex
from .saxtypes import ComplexArray1D as ComplexArray1D
from .saxtypes import ComplexArrayND as ComplexArrayND
from .saxtypes import Float as Float
from .saxtypes import FloatArray1D as FloatArray1D
from .saxtypes import FloatArrayND as FloatArrayND
from .saxtypes import Int as Int
from .saxtypes import IntArray1D as IntArray1D
from .saxtypes import IntArrayND as IntArrayND
from .saxtypes import Model as Model
from .saxtypes import ModelFactory as ModelFactory
from .saxtypes import PortCombination as PortCombination
from .saxtypes import PortMap as PortMap
from .saxtypes import SCoo as SCoo
from .saxtypes import SDense as SDense
from .saxtypes import SDict as SDict
from .saxtypes import Settings as Settings
from .saxtypes import SType as SType
from .saxtypes import is_complex as is_complex
from .saxtypes import is_complex_float as is_complex_float
from .saxtypes import is_float as is_float
from .saxtypes import is_mixedmode as is_mixedmode
from .saxtypes import is_model as is_model
from .saxtypes import is_model_factory as is_model_factory
from .saxtypes import is_multimode as is_multimode
from .saxtypes import is_scoo as is_scoo
from .saxtypes import is_sdense as is_sdense
from .saxtypes import is_sdict as is_sdict
from .saxtypes import is_singlemode as is_singlemode
from .saxtypes import modelfactory as modelfactory
from .saxtypes import scoo as scoo
from .saxtypes import sdense as sdense
from .saxtypes import sdict as sdict
from .saxtypes import validate_model as validate_model
from .utils import block_diag as block_diag
from .utils import cartesian_product as cartesian_product
from .utils import clean_string as clean_string
from .utils import copy_settings as copy_settings
from .utils import denormalize as denormalize
from .utils import flatten_dict as flatten_dict
from .utils import get_inputs_outputs as get_inputs_outputs
from .utils import get_port_combinations as get_port_combinations
from .utils import get_ports as get_ports
from .utils import get_settings as get_settings
from .utils import grouped_interp as grouped_interp
from .utils import merge_dicts as merge_dicts
from .utils import mode_combinations as mode_combinations
from .utils import normalization as normalization
from .utils import normalize as normalize
from .utils import reciprocal as reciprocal
from .utils import rename_params as rename_params
from .utils import rename_ports as rename_ports
from .utils import try_complex_float as try_complex_float
from .utils import unflatten_dict as unflatten_dict
from .utils import update_settings as update_settings
from .utils import validate_multimode as validate_multimode
from .utils import validate_not_mixedmode as validate_not_mixedmode
from .utils import validate_sdict as validate_sdict
from .utils import validate_settings as validate_settings



""" SAX Backends """

from __future__ import annotations

import warnings
from typing import Any, Dict

from ..netlist import Component
from ..saxtypes import Model, SType
from .additive import (
    analyze_circuit_additive,
    analyze_instances_additive,
    evaluate_circuit_additive,
)
from .filipsson_gunnar import (
    analyze_circuit_fg,
    analyze_instances_fg,
    evaluate_circuit_fg,
)
from .forward_only import (
    analyze_circuit_forward,
    analyze_instances_forward,
    evaluate_circuit_forward,
)

circuit_backends = {
    "fg": (
        analyze_instances_fg,
        analyze_circuit_fg,
        evaluate_circuit_fg,
    ),
    "filipsson_gunnar": (
        analyze_instances_fg,
        analyze_circuit_fg,
        evaluate_circuit_fg,
    ),
    "additive": (
        analyze_instances_additive,
        analyze_circuit_additive,
        evaluate_circuit_additive,
    ),
    "forward": (
        analyze_instances_forward,
        analyze_circuit_forward,
        evaluate_circuit_forward,
    ),
}

backend_map = {
    "fg": "filipsson_gunnar",
    "filipsson_gunnar": "filipsson_gunnar",
    "additive": "additive",
    "forward": "forward",
}

try:
    from .klu import analyze_circuit_klu, analyze_instances_klu, evaluate_circuit_klu

    circuit_backends["klu"] = (
        analyze_instances_klu,
        analyze_circuit_klu,
        evaluate_circuit_klu,
    )
    circuit_backends["default"] = (
        analyze_instances_klu,
        analyze_circuit_klu,
        evaluate_circuit_klu,
    )
    backend_map["klu"] = "klu"
    backend_map["default"] = "klu"
except ImportError:
    circuit_backends["default"] = (
        analyze_instances_fg,
        analyze_circuit_fg,
        evaluate_circuit_fg,
    )
    backend_map["default"] = "filipsson_gunnar"
    warnings.warn(
        "klujax not found. Please install klujax for "
        "better performance during circuit evaluation!"
    )


def analyze_instances(
    instances: Dict[str, Component],
    models: Dict[str, Model],
) -> Any:
    return circuit_backends["default"][0](instances, models)


def analyze_circuit(
    analyzed_instances: Any,
    connections: Dict[str, str],
    ports: Dict[str, str],
) -> Any:
    return circuit_backends["default"][1](analyzed_instances, connections, ports)


def evaluate_circuit(
    analyzed: Any,
    instances: Dict[str, SType],
) -> SType:
    return circuit_backends["default"][2](analyzed, instances)



""" SAX Multimode support """

from __future__ import annotations

from functools import wraps
from typing import Dict, Tuple, Union, cast, overload

import jax.numpy as jnp

from .saxtypes import (
    Model,
    SCoo,
    SDense,
    SDict,
    SType,
    _consolidate_sdense,
    is_model,
    is_multimode,
    is_scoo,
    is_sdense,
    is_sdict,
    is_singlemode,
)
from .utils import (
    block_diag,
    mode_combinations,
    validate_multimode,
    validate_not_mixedmode,
)


@overload
def multimode(S: Model, modes: Tuple[str, ...] = ("TE", "TM")) -> Model:
    ...


@overload
def multimode(S: SDict, modes: Tuple[str, ...] = ("TE", "TM")) -> SDict:
    ...


@overload
def multimode(S: SCoo, modes: Tuple[str, ...] = ("TE", "TM")) -> SCoo:
    ...


@overload
def multimode(S: SDense, modes: Tuple[str, ...] = ("TE", "TM")) -> SDense:
    ...


def multimode(
    S: Union[SType, Model], modes: Tuple[str, ...] = ("TE", "TM")
) -> Union[SType, Model]:
    """Convert a single mode model to a multimode model"""
    if is_model(S):
        model = cast(Model, S)

        @wraps(model)
        def new_model(**params):
            return multimode(model(**params), modes=modes)

        return cast(Model, new_model)

    S = cast(SType, S)

    validate_not_mixedmode(S)
    if is_multimode(S):
        validate_multimode(S, modes=modes)
        return S

    if is_sdict(S):
        return _multimode_sdict(cast(SDict, S), modes=modes)
    elif is_scoo(S):
        return _multimode_scoo(cast(SCoo, S), modes=modes)
    elif is_sdense(S):
        return _multimode_sdense(cast(SDense, S), modes=modes)
    else:
        raise ValueError("cannot convert to multimode. Unknown stype.")


def _multimode_sdict(sdict: SDict, modes: Tuple[str, ...] = ("TE", "TM")) -> SDict:
    multimode_sdict = {}
    _mode_combinations = mode_combinations(modes)
    for (p1, p2), value in sdict.items():
        for m1, m2 in _mode_combinations:
            multimode_sdict[f"{p1}@{m1}", f"{p2}@{m2}"] = value
    return multimode_sdict


def _multimode_scoo(scoo: SCoo, modes: Tuple[str, ...] = ("TE", "TM")) -> SCoo:
    Si, Sj, Sx, port_map = scoo
    num_ports = len(port_map)
    mode_map = (
        {mode: i for i, mode in enumerate(modes)}
        if not isinstance(modes, dict)
        else cast(Dict, modes)
    )

    _mode_combinations = mode_combinations(modes)

    Si_m = jnp.concatenate(
        [Si + mode_map[m] * num_ports for m, _ in _mode_combinations], -1
    )
    Sj_m = jnp.concatenate(
        [Sj + mode_map[m] * num_ports for _, m in _mode_combinations], -1
    )
    Sx_m = jnp.concatenate([Sx for _ in _mode_combinations], -1)
    port_map_m = {
        f"{port}@{mode}": idx + mode_map[mode] * num_ports
        for mode in modes
        for port, idx in port_map.items()
    }

    return Si_m, Sj_m, Sx_m, port_map_m


def _multimode_sdense(sdense, modes=("TE", "TM")):
    Sx, port_map = sdense
    num_ports = len(port_map)
    mode_map = (
        {mode: i for i, mode in enumerate(modes)}
        if not isinstance(modes, dict)
        else modes
    )

    Sx_m = block_diag(*(Sx for _ in modes))

    port_map_m = {
        f"{port}@{mode}": idx + mode_map[mode] * num_ports
        for mode in modes
        for port, idx in port_map.items()
    }

    return Sx_m, port_map_m


@overload
def singlemode(S: Model, mode: str = "TE") -> Model:
    ...


@overload
def singlemode(S: SDict, mode: str = "TE") -> SDict:
    ...


@overload
def singlemode(S: SCoo, mode: str = "TE") -> SCoo:
    ...


@overload
def singlemode(S: SDense, mode: str = "TE") -> SDense:
    ...


def singlemode(S: Union[SType, Model], mode: str = "TE") -> Union[SType, Model]:
    """Convert multimode model to a singlemode model"""
    if is_model(S):
        model = cast(Model, S)

        @wraps(model)
        def new_model(**params):
            return singlemode(model(**params), mode=mode)

        return cast(Model, new_model)

    S = cast(SType, S)

    validate_not_mixedmode(S)
    if is_singlemode(S):
        return S
    if is_sdict(S):
        return _singlemode_sdict(cast(SDict, S), mode=mode)
    elif is_scoo(S):
        return _singlemode_scoo(cast(SCoo, S), mode=mode)
    elif is_sdense(S):
        return _singlemode_sdense(cast(SDense, S), mode=mode)
    else:
        raise ValueError("cannot convert to multimode. Unknown stype.")


def _singlemode_sdict(sdict: SDict, mode: str = "TE") -> SDict:
    singlemode_sdict = {}
    for (p1, p2), value in sdict.items():
        if p1.endswith(f"@{mode}") and p2.endswith(f"@{mode}"):
            p1, _ = p1.split("@")
            p2, _ = p2.split("@")
            singlemode_sdict[p1, p2] = value
    return singlemode_sdict


def _singlemode_scoo(scoo: SCoo, mode: str = "TE") -> SCoo:
    Si, Sj, Sx, port_map = scoo
    # no need to touch the data...
    # just removing some ports from the port map should be enough
    port_map = {
        port.split("@")[0]: idx
        for port, idx in port_map.items()
        if port.endswith(f"@{mode}")
    }
    return Si, Sj, Sx, port_map


def _singlemode_sdense(sdense: SDense, mode: str = "TE") -> SDense:
    Sx, port_map = sdense
    # no need to touch the data...
    # just removing some ports from the port map should be enough
    port_map = {
        port.split("@")[0]: idx
        for port, idx in port_map.items()
        if port.endswith(f"@{mode}")
    }
    return _consolidate_sdense(Sx, port_map)
